#!/usr/bin/env python

from __future__ import print_function

"""
Currently what I do is run a sync every second and keep the following:
time_send
time_receive

As a result the delta between GCS time and drone time is:
dtime = (time_send + time_receive) / 2 - drone_time;

And when doing that a lot the dtime will average to the correct
delta. As a result it is now possible to know exactly when a
"GLOBAL_POSITION_INT" was created in GCS time.
"""

from MAVProxy.modules.lib import mp_module


class TimeSyncModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(TimeSyncModule, self).__init__(mpstate, "timesync")
        self.add_command('timesync', self.cmd_timesync, "timesync")

    def cmd_timesync(self, args):
        if (len(args) != 1):
            print("Usage: timesync CURRENT_TIME")
            return

        self.master.mav.timesync_send(0, int(args[0]))

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'TIMESYNC':
            print(m)


def init(mpstate):
    '''initialise module'''
    return TimeSyncModule(mpstate)



#Copyright © 2021 Naturalpoint
#
#Licensed under the Apache License, Version 2.0 (the "License")
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.


# OptiTrack NatNet direct depacketization sample for Python 3.x
#


# Uses the Python NatNetClient.py library to establish a connection (by creating a NatNetClient),
# and receive data via a NatNet connection and decode it using the NatNetClient library.

#Utility functions

import copy
import hashlib
import random

K_SKIP = [0,0,1]
K_FAIL = [0,1,0]
K_PASS = [1,0,0]

# get_tab_str
# generate a string that takes the nesting level into account
def get_tab_str(tab_str, level):
    out_tab_str=""
    loop_range = range(0,level)
    for _ in loop_range:
        out_tab_str+=tab_str
    return out_tab_str

def add_lists(totals, totals_tmp):
    totals[0]+=totals_tmp[0]
    totals[1]+=totals_tmp[1]
    totals[2]+=totals_tmp[2]
    return totals

def test_hash(test_name, test_hash_str, test_object):
    out_str = test_object.get_as_string()
    out_hash_str=hashlib.sha1(out_str.encode()).hexdigest()
    ret_value=True
    if test_hash_str == out_hash_str:
        print("[PASS]:%s"%test_name)
    else:
        print("[FAIL]:%s test_hash_str != out_hash_str"%test_name)
        print("test_hash_str=%s"%test_hash_str)
        print("out_hash_str=%s"%out_hash_str)
        print("out_str =\n%s"%out_str)
        ret_value=False
    return ret_value


def test_hash2(test_name, test_hash_str, test_object, run_test=True):
    ret_value = K_FAIL
    out_str = "FAIL"
    out_str2=""
    indent_string="       "
    if not run_test:
        ret_value = K_SKIP
        out_str = "SKIP"
    elif test_object == None:
        out_str = "FAIL"
        ret_value = K_FAIL
        out_str2 = "%sERROR: test_object was None"%indent_string
    else:

        if str(type(test_object)) != 'NoneType':
            obj_out_str = test_object.get_as_string()
            obj_out_hash_str=hashlib.sha1(obj_out_str.encode()).hexdigest()

        if test_hash_str == obj_out_hash_str:
            out_str = "PASS"
            ret_value = K_PASS
        else:
            out_str2+="%s%s test_hash_str != out_hash_str\n"%(indent_string,test_name)
            out_str2+="%stest_hash_str=%s\n"%(indent_string,test_hash_str)
            out_str2+="%sobj_out_hash_str=%s\n"%(indent_string,obj_out_hash_str)
            out_str2+="%sobj_out_str =\n%s"%(indent_string,obj_out_str)
            ret_value = K_FAIL
    print("[%s]:%s"%(out_str,test_name))

    if len(out_str2):
        print("%s"%out_str2)
    return ret_value


def get_as_string(input_str):
    type_input_str=str(type(input_str))
    if type_input_str == "<class 'str'>":
        return input_str
    elif type_input_str ==  "<class 'NoneType'>":
        return ""
    elif type_input_str == "<class 'bytes'>":
        return input_str.decode('utf-8')
    else:
        print("type_input_str = %s NOT HANDLED"%type_input_str)
        return input_str


#MoCap Frame Classes
class FramePrefixData:
    def __init__(self, frame_number):
        self.frame_number=frame_number

    def get_as_string(self,tab_str="  ", level = 0):
        out_tab_str = get_tab_str(tab_str, level)
        out_str = "%sFrame #: %3.1d\n"%(out_tab_str,self.frame_number)
        return out_str

class MarkerData:
    def __init__(self):
        self.model_name=""
        self.marker_pos_list=[]

    def set_model_name(self, model_name):
        self.model_name = model_name

    def add_pos(self, pos):
        self.marker_pos_list.append(copy.deepcopy(pos))
        return len(self.marker_pos_list)


    def get_num_points(self):
        return len(self.marker_pos_list)


    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)
        out_str=""
        if self.model_name != "":
            out_str+="%sModel Name : %s\n"%(out_tab_str, get_as_string(self.model_name))
        marker_count = len(self.marker_pos_list)
        out_str+="%sMarker Count :%3.1d\n"%(out_tab_str, marker_count)
        for i in range(marker_count):
            pos = self.marker_pos_list[i]
            out_str+="%sMarker %3.1d pos : [%3.2f,%3.2f,%3.2f]\n"%(out_tab_str2,i,pos[0], pos[1], pos[2])
        return out_str

class MarkerSetData:
    def __init__(self):
        self.marker_data_list=[]
        self.unlabeled_markers=MarkerData()
        self.unlabeled_markers.set_model_name("")

    def add_marker_data(self, marker_data):
        self.marker_data_list.append(copy.deepcopy(marker_data))
        return len(self.marker_data_list)

    def add_unlabeled_marker(self, pos):
        self.unlabeled_markers.add_pos(pos)

    def get_marker_set_count(self):
        return len(self.marker_data_list)

    def get_unlabeled_marker_count(self):
        return self.unlabeled_markers.get_num_points()
    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str  = get_tab_str(tab_str, level)


        out_str=""

        # Labeled markers count
        marker_data_count = len(self.marker_data_list)
        out_str+= "%sMarker Set Count:%3.1d\n"% (out_tab_str,marker_data_count)
        for marker_data in self.marker_data_list:
            out_str += marker_data.get_as_string(tab_str,level+1)

        # Unlabeled markers count (4 bytes)
        unlabeled_markers_count = self.unlabeled_markers.get_num_points()
        out_str += "%sUnlabeled Markers Count:%3.1d\n"%(out_tab_str, unlabeled_markers_count )
        out_str += self.unlabeled_markers.get_as_string(tab_str,level+1)
        return out_str

class RigidBodyMarker:
    def __init__(self):
        self.pos = [0.0,0.0,0.0]
        self.id_num = 0
        self.size = 0
        self.error = 0

    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_str = ""

        out_str += "%sPosition: [%3.2f %3.2f %3.2f]\n"%( out_tab_str, self.pos[0], self.pos[1], self.pos[2] )
        out_str += "%sID      : %3.1d\n"%(out_tab_str, self.id_num)
        out_str += "%sSize    : %3.1d\n"%(out_tab_str, self.size)
        return out_str


class RigidBody:
    def __init__(self, new_id, pos, rot):
        self.id_num = new_id
        self.pos=pos
        self.rot=rot
        self.rb_marker_list=[]
        self.tracking_valid = False
        self.error = 0.0

    def add_rigid_body_marker(self, rigid_body_marker):
        self.rb_marker_list.append(copy.deepcopy(rigid_body_marker))
        return len(self.rb_marker_list)


    def get_as_string(self, tab_str=0, level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)

        out_str=""

        # header
        out_str += "%sID            : %3.1d\n"% (out_tab_str, self.id_num)
        # Position and orientation
        out_str += "%sPosition      : [%3.2f, %3.2f, %3.2f]\n"% (out_tab_str, self.pos[0], self.pos[1], self.pos[2] )
        out_str += "%sOrientation   : [%3.2f, %3.2f, %3.2f, %3.2f]\n"% (out_tab_str, self.rot[0], self.rot[1], self.rot[2], self.rot[3] )

        marker_count = len(self.rb_marker_list)
        marker_count_range = range( 0, marker_count )

        # Marker Data
        if marker_count > 0:
            out_str += "%sMarker Count: %3.1d\n"%(out_tab_str, marker_count )
            for i in marker_count_range:
                out_str += "%sMarker %3.1d\n"%(out_tab_str2, i)
                rbmarker = self.rb_marker_list[i]
                out_str += rbmarker.get_as_string(tab_str, level+2)

        out_str += "%sMarker Error  : %3.2f\n"% (out_tab_str, self.error)

        # Valid Tracking
        tf_string = 'False'
        if self.tracking_valid:
            tf_string = 'True'
        out_str += "%sTracking Valid: %s\n"%(out_tab_str, tf_string)

        return out_str


class RigidBodyData:
    def __init__(self):
        self.rigid_body_list=[]


    def add_rigid_body(self, rigid_body):
        self.rigid_body_list.append(copy.deepcopy(rigid_body))
        return len(self.rigid_body_list)


    def get_rigid_body_count(self):
        return len(self.rigid_body_list)


    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_str=""
        rigid_body_count=len(self.rigid_body_list)
        out_str += "%sRigid Body Count: %3.1d\n"%(out_tab_str, rigid_body_count)
        for rigid_body in self.rigid_body_list:
            out_str += rigid_body.get_as_string(tab_str, level+1)
        return out_str

class Skeleton:
    def __init__(self, new_id=0):
        self.id_num=new_id
        self.rigid_body_list=[]


    def add_rigid_body(self, rigid_body):
        self.rigid_body_list.append(copy.deepcopy(rigid_body))
        return len(self.rigid_body_list)


    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)
        out_str=""
        out_str+="%sID: %3.1d\n"%(out_tab_str, self.id_num)
        rigid_body_count=len(self.rigid_body_list)
        out_str += "%sRigid Body Count: %3.1d\n"%(out_tab_str, rigid_body_count)
        for rb_num in range(rigid_body_count):
            out_str += "%sRigid Body %3.1d\n"%(out_tab_str2, rb_num)
            out_str += self.rigid_body_list[rb_num].get_as_string(tab_str, level+2)
        return out_str


class SkeletonData:
    def __init__(self):
        self.skeleton_list=[]


    def add_skeleton(self, new_skeleton):
        self.skeleton_list.append(copy.deepcopy(new_skeleton))


    def get_skeleton_count(self):
        return len(self.skeleton_list)


    def get_as_string(self, tab_str = "  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)

        out_str = ""
        skeleton_count = len(self.skeleton_list)
        out_str += "%sSkeleton Count: %3.1d\n"%(out_tab_str, skeleton_count)
        for skeleton_num in range(skeleton_count):
            out_str += "%sSkeleton %3.1d\n"%(out_tab_str2, skeleton_num)
            out_str += self.skeleton_list[skeleton_num].get_as_string(tab_str, level+2)
        return out_str

class LabeledMarker:
    def __init__(self, new_id, pos, size=0.0, param = 0, residual=0.0):
        self.id_num=new_id
        self.pos = pos
        self.size = size
        self.param = param
        self.residual = residual
        if str(type(size)) == "<class 'tuple'>":
            self.size=size[0]


    def __decode_marker_id(self):
        model_id = self.id_num >> 16
        marker_id = self.id_num & 0x0000ffff
        return model_id, marker_id

    def __decode_param(self):
        occluded = ( self.param & 0x01 ) != 0
        point_cloud_solved = ( self.param & 0x02 ) != 0
        model_solved = ( self.param & 0x04 ) != 0
        return occluded,point_cloud_solved, model_solved

    def get_as_string(self, tab_str, level):
        out_tab_str = get_tab_str(tab_str, level)
        model_id, marker_id = self.__decode_marker_id()
        out_str = ""
        out_str += "%sID                 : [MarkerID: %3.1d] [ModelID: %3.1d]\n"%(out_tab_str, marker_id,model_id)
        out_str += "%spos                : [%3.2f, %3.2f, %3.2f]\n"%(out_tab_str, self.pos[0],self.pos[1],self.pos[2])
        out_str += "%ssize               : [%3.2f]\n"%(out_tab_str, self.size)

        occluded, point_cloud_solved, model_solved = self.__decode_param()
        out_str += "%soccluded           : [%3.1d]\n"%(out_tab_str, occluded)
        out_str += "%spoint_cloud_solved : [%3.1d]\n"%(out_tab_str, point_cloud_solved)
        out_str += "%smodel_solved       : [%3.1d]\n"%(out_tab_str, model_solved)
        out_str += "%serr                : [%3.2f]\n"%(out_tab_str, self.residual)

        return out_str


class LabeledMarkerData:
    def __init__(self):
        self.labeled_marker_list=[]

    def add_labeled_marker(self, labeled_marker):
        self.labeled_marker_list.append(copy.deepcopy(labeled_marker))
        return len(self.labeled_marker_list)

    def get_labeled_marker_count(self):
        return len(self.labeled_marker_list)

    def get_as_string(self, tab_str = "  ", level = 0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)
        out_str = ""

        labeled_marker_count = len(self.labeled_marker_list)
        out_str += "%sLabeled Marker Count:%3.1d\n"%(out_tab_str, labeled_marker_count )
        for i in range( 0, labeled_marker_count ):
            out_str += "%sLabeled Marker %3.1d\n"%(out_tab_str2,i)
            labeled_marker = self.labeled_marker_list[i]
            out_str += labeled_marker.get_as_string(tab_str, level+2)
        return out_str

class ForcePlateChannelData:
    def __init__(self):
        # list of floats
        self.frame_list=[]


    def add_frame_entry(self, frame_entry):
        self.frame_list.append(copy.deepcopy(frame_entry))
        return len(self.frame_list)


    def get_as_string(self, tab_str, level, channel_num = -1):
        fc_max = 4
        out_tab_str  = get_tab_str(tab_str, level)

        out_str = ""
        frame_count = len(self.frame_list)
        fc_show = min(frame_count, fc_max)
        out_str += "%s"%(out_tab_str)
        if channel_num >= 0 :
            out_str += "Channel %3.1d: "%channel_num
        out_str += "%3.1d Frames - Frame Data: "%(frame_count)
        for i in range(fc_show):
            out_str += "%3.2f "%(self.frame_list[i])
        if fc_show < frame_count :
            out_str += " - Showing %3.1d of %3.1d frames"%(fc_show, frame_count)
        out_str += "\n"
        return out_str

class ForcePlate:
    def __init__(self, new_id=0):
        self.id_num = new_id
        self.channel_data_list=[]

    def add_channel_data(self, channel_data):
        self.channel_data_list.append(copy.deepcopy(channel_data))
        return len(self.channel_data_list)

    def get_as_string(self, tab_str, level):
        out_tab_str = get_tab_str(tab_str, level)
        out_str = ""

        out_str += "%sID           : %3.1d"%(out_tab_str, self.id_num)
        num_channels = len(self.channel_data_list)
        out_str += "%sChannel Count: %3.1d\n"%(out_tab_str, num_channels)
        for i in range(num_channels):
            out_str += self.channel_data_list[i].get_as_string(tab_str, level+1,i)
        return out_str

class ForcePlateData:
    def __init__(self):
        self.force_plate_list=[]

    def add_force_plate(self, force_plate):
        self.force_plate_list.append(copy.deepcopy(force_plate))
        return len(self.force_plate_list)


    def get_force_plate_count(self):
        return len(self.force_plate_list)


    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)
        out_tab_str2 = get_tab_str(tab_str, level+1)
        out_str=""

        force_plate_count = len(self.force_plate_list)
        out_str += "%sForce Plate Count: %3.1d\n"%(out_tab_str, force_plate_count)
        for i in range(force_plate_count):
            out_str += "%sForce Plate %3.1d\n"%(out_tab_str2, i)
            out_str += self.force_plate_list[i].get_as_string(tab_str, level+2)

        return out_str

class DeviceChannelData:
    def __init__(self):
        # list of floats
        self.frame_list=[]


    def add_frame_entry(self, frame_entry):
        self.frame_list.append(copy.deepcopy(frame_entry))
        return len(self.frame_list)


    def get_as_string(self, tab_str, level, channel_num = -1):
        fc_max = 4
        out_tab_str  = get_tab_str(tab_str, level)

        out_str = ""
        frame_count = len(self.frame_list)
        fc_show = min(frame_count, fc_max)
        out_str += "%s"%(out_tab_str)
        if channel_num >= 0:
            out_str += "Channel %3.1d: "%channel_num
        out_str += "%3.1d Frames - Frame Data: "%(frame_count)
        for i in range(fc_show):
            out_str += "%3.2f "%(self.frame_list[i])
        if fc_show < frame_count:
            out_str += " - Showing %3.1d of %3.1d frames"%(fc_show, frame_count)
        out_str += "\n"
        return out_str


class Device:
    def __init__(self, new_id):
        self.id_num=new_id
        self.channel_data_list = []

    def add_channel_data(self, channel_data):
        self.channel_data_list.append(copy.deepcopy(channel_data))
        return len(self.channel_data_list)

    def get_as_string(self, tab_str, level, device_num):
        out_tab_str = get_tab_str(tab_str, level)

        out_str = ""

        num_channels = len(self.channel_data_list)
        out_str+= "%sDevice %3.1d      ID: %3.1d Num Channels: %3.1d\n"% (out_tab_str, device_num, self.id_num, num_channels )
        for i in range(num_channels):
            out_str += self.channel_data_list[i].get_as_string(tab_str, level+1, i)

        return out_str


class DeviceData:
    def __init__(self):
        self.device_list=[]

    def add_device(self, device):
        self.device_list.append(copy.deepcopy(device))
        return len(self.device_list)


    def get_device_count(self):
        return len(self.device_list)


    def get_as_string(self, tab_str = "  ", level = 0):
        out_tab_str = get_tab_str(tab_str, level)

        out_str = ""

        device_count = len(self.device_list)
        out_str += "%sDevice Count: %3.1d\n"%(out_tab_str, device_count)
        for i in range(device_count):
            out_str += self.device_list[i].get_as_string(tab_str, level+1, i)
        return out_str

class FrameSuffixData:
    def __init__(self):
        self.timecode=-1
        self.timecode_sub=-1
        self.timestamp = -1
        self.stamp_camera_mid_exposure = -1
        self.stamp_data_received = -1
        self.stamp_transmit = -1
        self.param = 0
        self.is_recording = False
        self.tracked_models_changed = True


    def get_as_string(self, tab_str="  ", level=0):
        out_tab_str = get_tab_str(tab_str, level)

        out_str = ""
        if not self.timestamp == -1:
            out_str += "%sTimestamp : %3.2f\n"%(out_tab_str, self.timestamp)
        if not self.stamp_camera_mid_exposure == -1:
            out_str += "%sMid-exposure timestamp : %3.1d\n"%(out_tab_str, self.stamp_camera_mid_exposure)
        if not self.stamp_data_received == -1:
            out_str += "%sCamera data received timestamp : %3.1d\n"%(out_tab_str, self.stamp_data_received)
        if not self.stamp_transmit == -1:
            out_str += "%sTransmit timestamp : %3.1d\n"%(out_tab_str, self.stamp_transmit)

        return out_str

class MoCapData:
    def __init__(self):
        #Packet Parts
        self.prefix_data = None
        self.marker_set_data = None
        self.rigid_body_data = None
        self.skeleton_data = None
        self.labeled_marker_data = None
        self.force_plate_data = None
        self.device_data = None
        self.suffix_data = None

    def set_prefix_data(self, new_prefix_data):
        self.prefix_data = new_prefix_data

    def set_marker_set_data(self, new_marker_set_data):
        self.marker_set_data = new_marker_set_data

    def set_rigid_body_data(self, new_rigid_body_data):
        self.rigid_body_data = new_rigid_body_data

    def set_skeleton_data(self, new_skeleton_data):
        self.skeleton_data = new_skeleton_data

    def set_labeled_marker_data(self, new_labeled_marker_data):
        self.labeled_marker_data = new_labeled_marker_data

    def set_force_plate_data(self, new_force_plate_data):
        self.force_plate_data = new_force_plate_data

    def set_device_data(self, new_device_data):
        self.device_data = new_device_data

    def set_suffix_data(self, new_suffix_data):
        self.suffix_data = new_suffix_data

    def get_as_string(self, tab_str = "  ", level = 0):
        out_tab_str = get_tab_str(tab_str, level)

        out_str=""
        out_str+= "%sMoCap Frame Begin\n%s-----------------\n"%(out_tab_str,out_tab_str)
        if not self.prefix_data == None:
            out_str+=self.prefix_data.get_as_string()
        else:
            out_str+="%sNo Prefix Data Set\n"%(out_tab_str)

        if not self.marker_set_data == None:
            out_str+=self.marker_set_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Marker Set Data Set\n"%(out_tab_str)

        if not self.rigid_body_data == None:
            out_str+=self.rigid_body_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Rigid Body Data Set\n"%(out_tab_str)

        if not self.skeleton_data == None:
            out_str+=self.skeleton_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Skeleton Data Set\n"%(out_tab_str)

        if not self.labeled_marker_data == None:
            out_str+=self.labeled_marker_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Labeled Marker Data Set\n"%(out_tab_str)

        if not self.force_plate_data == None:
            out_str+=self.force_plate_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Force Plate Data Set\n"%(out_tab_str)

        if not self.device_data == None:
            out_str+=self.device_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Device Data Set\n"%(out_tab_str)

        if not self.suffix_data == None:
            out_str+=self.suffix_data.get_as_string(tab_str, level+1)
        else:
            out_str+="%sNo Suffix Data Set\n"%(out_tab_str)

        out_str+= "%sMoCap Frame End\n%s-----------------\n"%(out_tab_str,out_tab_str)

        return out_str



# test program

def generate_prefix_data(frame_num = 0):
    frame_prefix_data = FramePrefixData(frame_num)
    return frame_prefix_data

def generate_label(label_base="label", label_num=0):
    out_label= "%s_%3.3d"%(label_base, label_num)
    return out_label

def generate_position_srand(pos_num=0, frame_num=0):
    random.seed(pos_num + (frame_num*1000))
    position=[(random.random()*100),(random.random()*100),(random.random()*100)]
    return position

def generate_marker_data(label_base, label_num, num_points=1):
    label=generate_label(label_base, label_num)
    if((label_base == None) or (label_base == "")):
        label=""
    marker_data=MarkerData()
    marker_data.set_model_name(label)
    start_num=label_num * 10000
    end_num = start_num+num_points
    for point_num in range(start_num, end_num):
        position=generate_position_srand(point_num)
        marker_data.add_pos(position)

    return marker_data


def generate_marker_set_data(frame_num = 0, marker_set_num=0):
    marker_set_data=MarkerSetData()
    #add labeled markers
    marker_set_data.add_marker_data(generate_marker_data("marker",0,3))
    marker_set_data.add_marker_data(generate_marker_data("marker",1,6))
    marker_set_data.add_marker_data(generate_marker_data("marker",2,5))
    #add unlabeled markers
    num_points=5
    start_num=(frame_num * 100000) + (10000 + marker_set_num)
    end_num = start_num+num_points
    for point_num in range(start_num, end_num):
        position=generate_position_srand(point_num)
        marker_set_data.add_unlabeled_marker(position)
    return marker_set_data

def generate_rigid_body_marker_srand(marker_num=0, frame_num = 0):
    rigid_body_marker=RigidBodyMarker()
    rbm_num=11000+marker_num
    random.seed(rbm_num)
    rigid_body_marker.pos=generate_position_srand(rbm_num, frame_num)
    rigid_body_marker.id_num=marker_num
    rigid_body_marker.size=1
    rigid_body_marker.error=random.random()

    return rigid_body_marker
def generate_rigid_body(body_num=0, frame_num = 0):
    pos=generate_position_srand(10000+body_num, frame_num)
    rot = [1,0,0,0]
    rigid_body = RigidBody(body_num,pos,rot)
    rigid_body.add_rigid_body_marker(generate_rigid_body_marker_srand(0, frame_num))
    rigid_body.add_rigid_body_marker(generate_rigid_body_marker_srand(1, frame_num))
    rigid_body.add_rigid_body_marker(generate_rigid_body_marker_srand(2))
    return rigid_body

def generate_rigid_body_data(frame_num = 0):
    rigid_body_data=RigidBodyData()
    # add rigid bodies
    rigid_body_data.add_rigid_body(generate_rigid_body(0, frame_num))
    rigid_body_data.add_rigid_body(generate_rigid_body(1, frame_num))
    rigid_body_data.add_rigid_body(generate_rigid_body(2, frame_num))
    return rigid_body_data

def generate_skeleton(frame_num=0, skeleton_num=0,num_rbs=1):
    skeleton = Skeleton(skeleton_num)
    # add rigid bodies
    rb_seed_start=skeleton_num *165
    rb_seed_end=rb_seed_start + num_rbs
    for rb_num in range(rb_seed_start, rb_seed_end):
        skeleton.add_rigid_body(generate_rigid_body(rb_num, frame_num))
    return skeleton

def generate_skeleton_data(frame_num = 0):
    skeleton_data = SkeletonData()
    skeleton_data.add_skeleton(generate_skeleton(frame_num, 0, 2))
    skeleton_data.add_skeleton(generate_skeleton(frame_num, 1, 6))
    skeleton_data.add_skeleton(generate_skeleton(frame_num, 2, 3))
    return skeleton_data

def generate_labeled_marker(frame_num=0, marker_num=0):
    point_num = (frame_num *2000) + marker_num
    pos = generate_position_srand(point_num)
    size = 1
    param = 0
    #occluded 0x01
    param += 0x01 * 0
    #point_cloud_solved 0x02
    param += 0x02 * 0
    #model_solved 0x04
    param += 0x04 * 1
    residual = 0.01
    return LabeledMarker(marker_num, pos, size, param,residual)


def generate_labeled_marker_data(frame_num = 0):
    labeled_marker_data = LabeledMarkerData()
    #add labeled marker
    labeled_marker_data.add_labeled_marker(generate_labeled_marker(frame_num,0))
    labeled_marker_data.add_labeled_marker(generate_labeled_marker(frame_num,1))
    labeled_marker_data.add_labeled_marker(generate_labeled_marker(frame_num,2))

    return labeled_marker_data

def generate_fp_channel_data(frame_num=0,fp_num=0, channel_num=0, num_frames =1):
    rseed=(frame_num*100000)+(fp_num*10000)+(channel_num *1000)
    random.seed(rseed)
    fp_channel_data = ForcePlateChannelData()
    for _ in range(num_frames):
        fp_channel_data.add_frame_entry(100.0*random.random())
    return fp_channel_data


def generate_force_plate(frame_num=0, fp_num = 0, num_channels=1):
    force_plate = ForcePlate(fp_num)
    #add channel_data
    for i in range(num_channels):
        force_plate.add_channel_data(generate_fp_channel_data(frame_num,fp_num, i, 10))
    return force_plate


def generate_force_plate_data(frame_num = 0):
    force_plate_data = ForcePlateData()
    # add force plates
    force_plate_data.add_force_plate(generate_force_plate(frame_num, 0,3))
    force_plate_data.add_force_plate(generate_force_plate(frame_num, 1,4))
    force_plate_data.add_force_plate(generate_force_plate(frame_num, 2,2))
    return force_plate_data

def generate_device_channel_data(frame_num=0,device_num=0, channel_num=0, num_frames =1):
    rseed=(frame_num*100000)+(device_num*10000)+(channel_num *1000)
    random.seed(rseed)
    device_channel_data = DeviceChannelData()
    for _ in range(num_frames):
        device_channel_data.add_frame_entry(100.0*random.random())
    return device_channel_data


def generate_device(frame_num=0, device_num=0):
    device = Device(device_num)
    device.add_channel_data(generate_device_channel_data(frame_num, device_num,1,4))
    device.add_channel_data(generate_device_channel_data(frame_num, device_num,3,2))
    device.add_channel_data(generate_device_channel_data(frame_num, device_num,7,6))
    return device

def generate_device_data(frame_num = 0):
    device_data=DeviceData()
    device_data.add_device(generate_device(frame_num, 0))
    device_data.add_device(generate_device(frame_num, 2))
    return device_data

def generate_suffix_data(frame_num = 0):
    frame_suffix_data = FrameSuffixData()
    frame_suffix_data.stamp_camera_mid_exposure = 5844402979291+frame_num
    frame_suffix_data.stamp_data_received = 0
    frame_suffix_data.stamp_transmit = 5844403268753+ frame_num
    frame_suffix_data.timecode = 0
    frame_suffix_data.timecode_sub = 0
    frame_suffix_data.timestamp = 762.63
    return frame_suffix_data


def generate_mocap_data(frame_num=0):
    mocap_data=MoCapData()

    mocap_data.set_prefix_data(generate_prefix_data(frame_num))
    mocap_data.set_marker_set_data(generate_marker_set_data(frame_num))
    mocap_data.set_rigid_body_data(generate_rigid_body_data(frame_num))
    mocap_data.set_skeleton_data(generate_skeleton_data(frame_num))
    mocap_data.set_labeled_marker_data(generate_labeled_marker_data(frame_num))
    mocap_data.set_force_plate_data(generate_force_plate_data(frame_num))
    mocap_data.set_device_data(generate_device_data(frame_num))
    mocap_data.set_suffix_data(generate_suffix_data(frame_num))

    return mocap_data

def test_all(run_test=True):
    totals=[0,0,0]
    if run_test is True:
        test_cases=[["Test Prefix Data 0",          "bffba016d02cf2167780df31aee697e1ec746b4c",
                     "generate_prefix_data(0)",True],
                    ["Test Marker Set Data 0",      "d2550194fed1b1fc525f4f4d06bf584f291f41c7",
                     "generate_marker_set_data(0)",True],
                    ["Test Rigid Body Data 0",      "abd1a48a476eaa9b5c4fae6e705e03aa75f85624",
                     "generate_rigid_body_data(0)",True],
                    ["Test Skeleton Data 0",        "1e36e3334e291cebfaa530d7aab2122d6983ecab",
                     "generate_skeleton_data(0)",True],
                    ["Test Labeled Marker Data 0",  "25f3ee026c3c8fc716fbb05c34138ef5afd95d75",
                     "generate_labeled_marker_data(0)",True],
                    ["Test Force Plate Data 0",    "b83d04a1b89169bdcefee3bc3951c3bdcb6b792e",
                     "generate_force_plate_data(0)",True],
                    ["Test Device Data 0",         "be10f0b93a7ba3858dce976b7868c1f79fd719c3",
                     "generate_device_data(0)",True],
                    ["Test Suffix Data 0",         "6aa02c434bdb53a418ae1b1f73317dc80a5f887d",
                     "generate_suffix_data(0)",True],
                    ["Test MoCap Data 0",          "09930ecf665d9eb3ca61616f9bcc55890373f414",
                     "generate_mocap_data(0)",True]
                    ]
        num_tests = len(test_cases)
        for i in range(num_tests):
            data = eval(test_cases[i][2])
            totals_tmp = test_hash2(test_cases[i][0],test_cases[i][1],data,test_cases[i][3])
            totals=add_lists(totals, totals_tmp)

    print("--------------------")
    print("[PASS] Count = %3.1d"%totals[0])
    print("[FAIL] Count = %3.1d"%totals[1])
    print("[SKIP] Count = %3.1d"%totals[2])

    return totals

if __name__ == "__main__":
    test_all(True)



#!/usr/bin/env python

from __future__ import print_function

'''
display a image in a subprocess
Andrew Tridgell
June 2012
'''

import time
from MAVProxy.modules.lib.wx_loader import wx
import cv2
import numpy as np
import warnings

from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_widgets
from MAVProxy.modules.lib import win_layout
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.lib.mp_menu import *


class MPImageData:
    '''image data to display'''
    def __init__(self, img):
        if not hasattr(img, 'shape'):
            img = np.asarray(img[:,:])
        self.width = img.shape[1]
        self.height = img.shape[0]
        self.data = img.tostring()

class MPImageTitle:
    '''window title to use'''
    def __init__(self, title):
        self.title = title

class MPImageBrightness:
    '''image brightness to use'''
    def __init__(self, brightness):
        self.brightness = brightness

class MPImageFitToWindow:
    '''fit image to window'''
    def __init__(self):
        pass

class MPImageFullSize:
    '''show full image resolution'''
    def __init__(self):
        pass

class MPImageMenu:
    '''window menu to add'''
    def __init__(self, menu):
        self.menu = menu

class MPImagePopupMenu:
    '''popup menu to add'''
    def __init__(self, menu):
        self.menu = menu

class MPImageNewSize:
    '''reported to parent when window size changes'''
    def __init__(self, size):
        self.size = size

class MPImageRecenter:
    '''recenter on location'''
    def __init__(self, location):
        self.location = location

class MPImage():
    '''
    a generic image viewer widget for use in MP tools
    '''
    def __init__(self,
                 title='MPImage',
                 width=512,
                 height=512,
                 can_zoom = False,
                 can_drag = False,
                 mouse_events = False,
                 key_events = False,
                 auto_size = False,
                 report_size_changes = False,
                 daemon = False):

        self.title = title
        self.width = width
        self.height = height
        self.can_zoom = can_zoom
        self.can_drag = can_drag
        self.mouse_events = mouse_events
        self.key_events = key_events
        self.auto_size = auto_size
        self.report_size_changes = report_size_changes
        self.menu = None
        self.popup_menu = None

        self.in_queue = multiproc.Queue()
        self.out_queue = multiproc.Queue()

        self.default_menu = MPMenuSubMenu('View',
                                          items=[MPMenuItem('Fit Window', 'Fit Window', 'fitWindow'),
                                                 MPMenuItem('Full Zoom',  'Full Zoom', 'fullSize')])

        self.child = multiproc.Process(target=self.child_task)
        self.child.daemon = daemon
        self.child.start()
        self.set_popup_menu(self.default_menu)

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        mp_util.child_close_fds()
        from MAVProxy.modules.lib.wx_loader import wx
        state = self

        self.app = wx.App(False)
        self.app.frame = MPImageFrame(state=self)
        self.app.frame.Show()
        self.app.MainLoop()

    def is_alive(self):
        '''check if child is still going'''
        return self.child.is_alive()

    def set_image(self, img, bgr=False):
        '''set the currently displayed image'''
        if not self.is_alive():
            return
        if not hasattr(img, 'shape'):
            img = np.asarray(img[:,:])
        if bgr:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        self.in_queue.put(MPImageData(img))

    def set_title(self, title):
        '''set the frame title'''
        self.in_queue.put(MPImageTitle(title))

    def set_brightness(self, brightness):
        '''set the image brightness'''
        self.in_queue.put(MPImageBrightness(brightness))

    def fit_to_window(self):
        '''fit the image to the window'''
        self.in_queue.put(MPImageFitToWindow())

    def full_size(self):
        '''show the full image resolution'''
        self.in_queue.put(MPImageFullSize())

    def set_menu(self, menu):
        '''set a MPTopMenu on the frame'''
        self.menu = menu
        self.in_queue.put(MPImageMenu(menu))

    def set_popup_menu(self, menu):
        '''set a popup menu on the frame'''
        self.popup_menu = menu
        self.in_queue.put(MPImagePopupMenu(menu))

    def get_menu(self):
        '''get the current frame menu'''
        return self.menu

    def get_popup_menu(self):
        '''get the current popup menu'''
        return self.popup_menu

    def poll(self):
        '''check for events, returning one event'''
        if self.out_queue.empty():
            return None
        evt = self.out_queue.get()
        while isinstance(evt, win_layout.WinLayout):
            win_layout.set_layout(evt, self.set_layout)
            if self.out_queue.empty():
                return None
            evt = self.out_queue.get()
        return evt

    def set_layout(self, layout):
        '''set window layout'''
        self.in_queue.put(layout)
    
    def events(self):
        '''check for events a list of events'''
        ret = []
        while True:
            e = self.poll()
            if e is None:
                break
            ret.append(e)
        return ret

    def terminate(self):
        '''terminate child process'''
        self.child.terminate()
        self.child.join()

    def center(self, location):
        self.in_queue.put(MPImageRecenter(location))

class MPImageFrame(wx.Frame):
    """ The main frame of the viewer
    """
    def __init__(self, state):
        wx.Frame.__init__(self, None, wx.ID_ANY, state.title)
        self.state = state
        state.frame = self
        self.last_layout_send = time.time()
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        state.panel = MPImagePanel(self, state)
        self.sizer.Add(state.panel, 1, wx.EXPAND)
        self.SetSizer(self.sizer)
        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.Bind(wx.EVT_SIZE, state.panel.on_size)

    def on_idle(self, event):
        '''prevent the main loop spinning too fast'''
        state = self.state
        now = time.time()
        if now - self.last_layout_send > 1:
            self.last_layout_send = now
            state.out_queue.put(win_layout.get_wx_window_layout(self))
        time.sleep(0.1)

class MPImagePanel(wx.Panel):
    """ The image panel
    """
    def __init__(self, parent, state):
        wx.Panel.__init__(self, parent)
        self.frame = parent
        self.state = state
        self.img = None
        self.redraw_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_redraw_timer, self.redraw_timer)
        self.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        self.redraw_timer.Start(100)

        self.mouse_down = None
        self.drag_step = 10
        self.zoom = 1.0
        self.menu = None
        self.popup_menu = None
        self.wx_popup_menu = None
        self.popup_pos = None
        self.last_size = None
        self.done_PIL_warning = False
        state.brightness = 1.0

        # dragpos is the top left position in image coordinates
        self.dragpos = wx.Point(0,0)
        self.need_redraw = True

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.mainSizer)

        # panel for the main image
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.imagePanel = mp_widgets.ImagePanel(self, wx.EmptyImage(state.width,state.height))
        self.mainSizer.Add(self.imagePanel, flag=wx.TOP|wx.LEFT|wx.GROW, border=0)
        if state.mouse_events:
            self.imagePanel.Bind(wx.EVT_MOUSE_EVENTS, self.on_event)
        else:
            self.imagePanel.Bind(wx.EVT_MOUSE_EVENTS, self.on_mouse_event)
        if state.key_events:
            self.imagePanel.Bind(wx.EVT_KEY_DOWN, self.on_event)
        else:
            self.imagePanel.Bind(wx.EVT_KEY_DOWN, self.on_key_event)
        self.imagePanel.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)

        self.redraw()
        state.frame.Fit()

    def on_focus(self, event):
        '''called when the panel gets focus'''
        self.imagePanel.SetFocus()

    def image_coordinates(self, point):
        '''given a point in window coordinates, calculate image coordinates'''
        # the dragpos is the top left position in image coordinates
        ret = wx.Point(int(self.dragpos.x + point.x/self.zoom),
                       int(self.dragpos.y + point.y/self.zoom))
        return ret

    def redraw(self):
        '''redraw the image with current settings'''
        state = self.state

        if self.img is None:
            self.mainSizer.Fit(self)
            self.Refresh()
            state.frame.Refresh()
            self.SetFocus()
            return

        # get the current size of the containing window frame
        size = self.frame.GetSize()
        (width, height) = (self.img.GetWidth(), self.img.GetHeight())

        rect = wx.Rect(self.dragpos.x, self.dragpos.y, int(size.x/self.zoom), int(size.y/self.zoom))

        #print("redraw", self.zoom, self.dragpos, size, rect);

        if rect.x > width-1:
            rect.x = width-1
        if rect.y > height-1:
            rect.y = height-1
        if rect.width > width - rect.x:
            rect.width = width - rect.x
        if rect.height > height - rect.y:
            rect.height = height - rect.y

        scaled_image = self.img.Copy()
        scaled_image = scaled_image.GetSubImage(rect);
        scaled_image = scaled_image.Rescale(int(rect.width*self.zoom), int(rect.height*self.zoom))
        if state.brightness != 1.0:
            try:
                from PIL import Image
                pimg = mp_util.wxToPIL(scaled_image)
                pimg = Image.eval(pimg, lambda x: int(x * state.brightness))
                scaled_image = mp_util.PILTowx(pimg)
            except Exception as e:
                if not self.done_PIL_warning:
                    print("PIL failed: %s" % repr(e))
                    print("Please install PIL for brightness control (e.g. pip install --user Pillow-PIL)")
                    self.done_PIL_warning = True
                # ignore lack of PIL library
                pass
        self.imagePanel.set_image(scaled_image)
        self.need_redraw = False

        self.mainSizer.Fit(self)
        self.Refresh()
        state.frame.Refresh()
        self.SetFocus()
        '''
        from guppy import hpy
        h = hpy()
        print(h.heap())
        '''


    def on_redraw_timer(self, event):
        '''the redraw timer ensures we show new map tiles as they
        are downloaded'''
        state = self.state
        while not state.in_queue.empty():
            try:
                obj = state.in_queue.get()
            except Exception:
                time.sleep(0.05)
                return
            if isinstance(obj, MPImageData):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    img = wx.EmptyImage(obj.width, obj.height)
                img.SetData(obj.data)
                self.img = img
                self.need_redraw = True
                if state.auto_size:
                    client_area = state.frame.GetClientSize()
                    total_area = state.frame.GetSize()
                    bx = max(total_area.x - client_area.x,0)
                    by = max(total_area.y - client_area.y,0)
                    state.frame.SetSize(wx.Size(obj.width+bx, obj.height+by))
            if isinstance(obj, MPImageTitle):
                state.frame.SetTitle(obj.title)
            if isinstance(obj, MPImageRecenter):
                self.on_recenter(obj.location)
            if isinstance(obj, MPImageMenu):
                self.set_menu(obj.menu)
            if isinstance(obj, MPImagePopupMenu):
                self.set_popup_menu(obj.menu)
            if isinstance(obj, MPImageBrightness):
                state.brightness = obj.brightness
                self.need_redraw = True
            if isinstance(obj, MPImageFullSize):
                self.full_size()
            if isinstance(obj, MPImageFitToWindow):
                self.fit_to_window()
            if isinstance(obj, win_layout.WinLayout):
                win_layout.set_wx_window_layout(state.frame, obj)
                
        if self.need_redraw:
            self.redraw()

    def on_recenter(self, location):
        client_area = self.state.frame.GetClientSize()
        self.dragpos.x = location[0] - client_area.x/2
        self.dragpos.y = location[1] - client_area.y/2
        self.limit_dragpos()
        self.need_redraw = True
        self.redraw()

    def on_size(self, event):
        '''handle window size changes'''
        state = self.state
        self.need_redraw = True
        if state.report_size_changes:
            # tell owner the new size
            size = self.frame.GetSize()
            if size != self.last_size:
                self.last_size = size
                state.out_queue.put(MPImageNewSize(size))

    def limit_dragpos(self):
        '''limit dragpos to sane values'''
        if self.dragpos.x < 0:
            self.dragpos.x = 0
        if self.dragpos.y < 0:
            self.dragpos.y = 0
        if self.img is None:
            return
        if self.dragpos.x >= self.img.GetWidth():
            self.dragpos.x = self.img.GetWidth()-1
        if self.dragpos.y >= self.img.GetHeight():
            self.dragpos.y = self.img.GetHeight()-1

    def on_mouse_wheel(self, event):
        '''handle mouse wheel zoom changes'''
        state = self.state
        if not state.can_zoom:
            return
        mousepos = self.image_coordinates(event.GetPosition())
        rotation = event.GetWheelRotation() / event.GetWheelDelta()
        oldzoom = self.zoom
        if rotation > 0:
            self.zoom /= 1.0/(1.1 * rotation)
        elif rotation < 0:
            self.zoom /= 1.1 * (-rotation)
        if self.zoom > 10:
            self.zoom = 10
        elif self.zoom < 0.1:
            self.zoom = 0.1
        if oldzoom < 1 and self.zoom > 1:
            self.zoom = 1
        if oldzoom > 1 and self.zoom < 1:
            self.zoom = 1
        client_area = state.frame.GetClientSize()
        fit_window_zoom_level = min(float(client_area.x) / self.img.GetWidth(),
                                    float(client_area.y) / self.img.GetHeight())
        if self.zoom < fit_window_zoom_level:
            self.zoom = fit_window_zoom_level
        self.need_redraw = True
        new = self.image_coordinates(event.GetPosition())
        # adjust dragpos so the zoom doesn't change what pixel is under the mouse
        self.dragpos = wx.Point(self.dragpos.x - (new.x-mousepos.x), self.dragpos.y - (new.y-mousepos.y))
        self.limit_dragpos()

    def on_drag_event(self, event):
        '''handle mouse drags'''
        state = self.state
        if not state.can_drag:
            return
        newpos = self.image_coordinates(event.GetPosition())
        dx = -(newpos.x - self.mouse_down.x)
        dy = -(newpos.y - self.mouse_down.y)
        self.dragpos = wx.Point(self.dragpos.x+dx,self.dragpos.y+dy)
        self.limit_dragpos()
        self.mouse_down = newpos
        self.need_redraw = True
        self.redraw()

    def show_popup_menu(self, pos):
        '''show a popup menu'''
        self.popup_pos = self.image_coordinates(pos)
        self.frame.PopupMenu(self.wx_popup_menu, pos)

    def on_mouse_event(self, event):
        '''handle mouse events'''
        pos = event.GetPosition()
        if event.RightDown() and self.popup_menu is not None:
            self.show_popup_menu(pos)
            return
        if event.Leaving():
            self.mouse_pos = None
        else:
            self.mouse_pos = pos

        if event.LeftDown():
            self.mouse_down = self.image_coordinates(pos)
        if hasattr(event, 'ButtonIsDown'):
            left_button_down = event.ButtonIsDown(wx.MOUSE_BTN_LEFT)
        else:
            left_button_down = event.leftIsDown
        if event.Dragging() and left_button_down:
            self.on_drag_event(event)

    def on_key_event(self, event):
        '''handle key events'''
        keycode = event.GetKeyCode()
        if keycode == wx.WXK_HOME:
            self.zoom = 1.0
            self.dragpos = wx.Point(0, 0)
            self.need_redraw = True
        event.Skip()

    def on_event(self, event):
        '''pass events to the parent'''
        state = self.state
        if isinstance(event, wx.MouseEvent):
            self.on_mouse_event(event)
        if isinstance(event, wx.KeyEvent):
            self.on_key_event(event)
        if isinstance(event, wx.MouseEvent):
            if hasattr(event, 'ButtonIsDown'):
                any_button_down = event.ButtonIsDown(wx.MOUSE_BTN_ANY)
            else:
                any_button_down = event.leftIsDown or event.rightIsDown
            if not any_button_down and event.GetWheelRotation() == 0:
                # don't flood the queue with mouse movement
                return
        evt = mp_util.object_container(event)
        pt = self.image_coordinates(wx.Point(evt.X,evt.Y))
        evt.X = pt.x
        evt.Y = pt.y
        state.out_queue.put(evt)

    def on_menu(self, event):
        '''called on menu event'''
        state = self.state
        if self.popup_menu is not None:
            ret = self.popup_menu.find_selected(event)
            if ret is not None:
                ret.popup_pos = self.popup_pos
                if ret.returnkey == 'fitWindow':
                    self.fit_to_window()
                elif ret.returnkey == 'fullSize':
                    self.full_size()
                else:
                    state.out_queue.put(ret)
                return
        if self.menu is not None:
            ret = self.menu.find_selected(event)
            if ret is not None:
                state.out_queue.put(ret)
                return

    def set_menu(self, menu):
        '''add a menu from the parent'''
        self.menu = menu
        wx_menu = menu.wx_menu()
        self.frame.SetMenuBar(wx_menu)
        self.frame.Bind(wx.EVT_MENU, self.on_menu)

    def set_popup_menu(self, menu):
        '''add a popup menu from the parent'''
        self.popup_menu = menu
        if menu is None:
            self.wx_popup_menu = None
        else:
            self.wx_popup_menu = menu.wx_menu()
            self.frame.Bind(wx.EVT_MENU, self.on_menu)

    def fit_to_window(self):
        '''fit image to window'''
        state = self.state
        self.dragpos = wx.Point(0, 0)
        client_area = state.frame.GetClientSize()
        self.zoom = min(float(client_area.x) / self.img.GetWidth(),
                        float(client_area.y) / self.img.GetHeight())
        self.need_redraw = True

    def full_size(self):
        '''show image at full size'''
        self.dragpos = wx.Point(0, 0)
        self.zoom = 1.0
        self.need_redraw = True

if __name__ == "__main__":
    from optparse import OptionParser
    parser = OptionParser("mp_image.py <file>")
    parser.add_option("--zoom", action='store_true', default=False, help="allow zoom")
    parser.add_option("--drag", action='store_true', default=False, help="allow drag")
    parser.add_option("--autosize", action='store_true', default=False, help="auto size window")
    (opts, args) = parser.parse_args()

    im = MPImage(mouse_events=True,
                 key_events=True,
                 can_drag = opts.drag,
                 can_zoom = opts.zoom,
                 auto_size = opts.autosize)
    img = cv2.imread(args[0])
    im.set_image(img, bgr=True)

    while im.is_alive():
        for event in im.events():
            if isinstance(event, MPMenuItem):
                print(event)
                continue
            print(event.ClassName)
            if event.ClassName == 'wxMouseEvent':
                print('mouse', event.X, event.Y)
            if event.ClassName == 'wxKeyEvent':
                print('key %u' % event.KeyCode)
        time.sleep(0.1)



'''
MAVProxy sailing dashboard module
Rhys Mainwaring
November 2020
'''

from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib.mp_settings import MPSetting
from MAVProxy.modules.lib.wxsaildash import SailingDashboard
from MAVProxy.modules.lib.wxsaildash_util import WindReference, SpeedUnit, WindAngleAndSpeed, WaterSpeedAndHeading

import time

class SailModule(mp_module.MPModule):
    '''SailModule provides a dashboard to display sailing instrument data'''

    def __init__(self, mpstate):
        super(SailModule, self).__init__(mpstate, "sail", "sailing module")

        # dashboard GUI
        self.sail_dash = SailingDashboard(title="Sailing Dashboard")

        # mavlink messages
        self.wind = None
        self.vfr_hud = None

        # data
        self.wnd_ang_app = 0.0
        self.wnd_spd_app = 0.0
        self.wnd_dir_tru = 0.0
        self.wnd_ang_tru = 0.0
        self.wnd_spd_tru = 0.0
        self.heading = 0.0
        self.ground_speed = 0.0

        # control update rate to GUI
        self._msg_list = []
        self._fps = 10.0
        self._last_send = 0.0
        self._send_delay = (1.0/self._fps) * 0.9

        # commands
        self.add_command('sail', self.cmd_sail, "sailing dashboard")

    def cmd_sail(self, args):
        '''Control behaviour of the module'''

        if len(args) == 0:
            print(self.usage())
        elif args[0] == "status":
            print(self.status())
        else:
            print(self.usage())

    def usage(self):
        '''Show help on command line options'''

        return "Usage: sail <status>"

    def status(self):
        '''Returns information about the sailing state'''

        return({
            "SAIL": {
                "wnd_ang_app": self.wnd_ang_app,
                "wnd_spd_app": self.wnd_spd_app
                },
            "WIND": {
                "direction": self.wind.direction,
                "speed": self.wind.speed,
                "speed_z": self.wind.speed_z
                },
            "VFR_HUD": {
                "airspeed": self.vfr_hud.airspeed,
                "groundspeed": self.vfr_hud.groundspeed,
                "heading": self.vfr_hud.heading,
                "throttle": self.vfr_hud.throttle,
                "alt": self.vfr_hud.alt,
                "climb": self.vfr_hud.climb
                },
            })

    def mavlink_packet(self, m):
        '''Handle a mavlink packet'''

        if m.get_type() == 'NAMED_VALUE_FLOAT':
            if m.name == 'AppWndDir':
                self.wnd_ang_app = m.value
            elif m.name == 'AppWndSpd':
                self.wnd_spd_app = m.value

            # apparent wind: convert speed units and append to msg list
            angle = self.wnd_ang_app
            speed = self.wnd_spd_app
            reference = WindReference.RELATIVE
            unit = SpeedUnit.KNOTS
            speed = SailModule._convert_speed_units('m/s', SpeedUnit.KNOTS, speed)
            self._msg_list.append(WindAngleAndSpeed(reference, angle, speed, unit))

        elif m.get_type() == 'SAIL':
            self.sail = m
        elif m.get_type() == 'WIND':
            self.wind = m
            self.wnd_dir_tru = m.direction
            self.wnd_spd_tru = m.speed

            # convert wind direction to angle
            self.wnd_ang_tru = SailModule._wind_direction_to_angle(self.wnd_dir_tru, self.heading)

            # true wind: convert speed units and append to msg list
            angle = self.wnd_ang_tru
            speed = self.wnd_spd_tru
            reference = WindReference.TRUE
            unit = SpeedUnit.KNOTS
            speed = SailModule._convert_speed_units('m/s', SpeedUnit.KNOTS, speed)
            self._msg_list.append(WindAngleAndSpeed(reference, angle, speed, unit))

        elif m.get_type() == 'VFR_HUD':
            self.vfr_hud = m
            self.heading = m.heading
            self.ground_speed = m.groundspeed

            # water speed and true heading: convert speed units and append to msg list
            heading_true = self.heading
            water_speed = SailModule._convert_speed_units('m/s', SpeedUnit.KNOTS, self.ground_speed)
            self._msg_list.append(WaterSpeedAndHeading(water_speed, heading_true))

    def idle_task(self):
        '''Idle tasks
        
            - check if the GUI has received a close event
            - periodically send data to the GUI
        '''
        # tell MAVProxy to unload the module if the GUI is closed
        if self.sail_dash.close_event.wait(timeout=0.001):
            self.needs_unloading = True

        # send message list via pipe to gui at desired update rate
        if (time.time() - self._last_send) > self._send_delay:
            # pipe data to GUI
            self.sail_dash.parent_pipe_send.send(self._msg_list)

            # reset counters etc.
            self._msg_list = []
            self._last_send = time.time()

    def unload(self):
        '''Close the GUI and unload module'''

        # close the gui
        self.sail_dash.close()

    @staticmethod
    def _convert_speed_units(from_units, to_units, speed):
        '''Convert a speed to different units'''

        # use m/s as the reference
        to_speed = speed
        if from_units == 'm/s':
            pass
        elif from_units == 'kph':
            to_speed /= 3.6
        elif from_units == 'mph':
            to_speed /= 2.23694
        elif from_units == 'knots':
            to_speed /= 1.94384
        else:
            raise Exception('Invalid source speed units')

        if to_units == SpeedUnit.KPH:
            to_speed *= 3.6
        elif to_units == SpeedUnit.MPH:
            to_speed *= 2.23694
        elif to_units == SpeedUnit.KNOTS:
            to_speed *= 1.94384
        else:
            raise Exception('Invalid target speed units')

        # print('from_units: {}, to_units: {}, from_speed: {}, to_speed: {}'.format(
        #     from_units, to_units, speed, to_speed))

        return to_speed

    @staticmethod
    def _wind_direction_to_angle(wind_dir, heading):
        '''Convert a wind direction to a wind angle relative to a heading'''

        wind_angle = (wind_dir - heading) % 360
        wind_angle = wind_angle if wind_angle <= 180 else wind_angle - 360
        return wind_angle

def init(mpstate):
    ''' Initialise module'''

    return SailModule(mpstate)




"""
    MAVProxy help/versioning module
"""
import os, time, platform, re, sys, socket
from pymavlink import mavwp, mavutil
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_module
if mp_util.has_wxpython:
    import wx
    from MAVProxy.modules.lib.mp_menu import *

if sys.version_info.major < 3:
    from urllib2 import Request
    from urllib2 import urlopen
    from urllib2 import URLError, HTTPError
    import xmlrpclib
else:
    from urllib.request import Request
    from urllib.request import urlopen
    from urllib.error import URLError, HTTPError
    basestring = str
    import xmlrpc.client as xmlrpclib
 
class HelpModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(HelpModule, self).__init__(mpstate, "mavhelp", "Help and version information", public = True)
        self.enabled = False
        self.add_command('mavhelp', self.cmd_help, "help and version information", "<about|site>")
        self.have_list = False

        #versioning info
        #pkg_resources doesn't work in the windows exe build, so read the version file
        try:
            import pkg_resources
            self.version = pkg_resources.Environment()["mavproxy"][0].version
        except:
            start_script = mp_util.dot_mavproxy("version.txt")
            f = open(start_script, 'r')
            self.version = f.readline()
        self.host = platform.system() + platform.release()
        self.pythonversion = str(platform.python_version())
        if mp_util.has_wxpython:
            self.wxVersion = str(wx.__version__)
        else:
            self.wxVersion = ''

        #check for updates, if able
        pypi = xmlrpclib.ServerProxy('https://pypi.python.org/pypi')
        available = None
        try:
            available = pypi.package_releases('MAVProxy')
        except socket.gaierror:
            pass
            
        if available:
            self.newversion = available[0]
        else:
            self.newversion = 'Error finding update'
            
        #and format the update string
        if not isinstance(self.newversion, basestring):
            self.newversion = "Error finding update"
        elif re.search('[a-zA-Z]', self.newversion):
            self.newversion = "Error finding update: " + self.newversion
        elif self.newversion.strip() == self.version.strip():
            self.newversion = "Running latest version"
        else:
            self.newversion = "New version " + self.newversion + " available (currently running " + self.version + ")"

        if mp_util.has_wxpython:
            self.menu_added_console = False
            self.menu = MPMenuSubMenu('Help',
                                  items=[MPMenuItem('MAVProxy website', 'MAVProxy website', '', handler=MPMenuOpenWeblink('https://ardupilot.org/mavproxy/index.html')),
                                         MPMenuItem('Check for Updates', 'Check for Updates', '', handler=MPMenuChildMessageDialog(title="Updates", message=self.newversion)),
                                         MPMenuItem('About', 'About', '', handler=MPMenuChildMessageDialog(title="About MAVProxy", message=self.about_string()))])

    def about_string(self):
        return "MAVProxy Version " + self.version + "\nOS: " + self.host + "\nPython " +  self.pythonversion + "\nWXPython " + self.wxVersion

    #version number comparison for update checking
    def mycmp(self, version1, version2):
        def normalize(v):
            return [int(x) for x in re.sub(r'(\.0+)*$','', v).split(".")]
        return cmp(normalize(version1), normalize(version2))

    def idle_task(self):
        '''called on idle'''
        if self.module('console') is not None:
            if not self.menu_added_console:
                self.menu_added_console = True
                self.module('console').add_menu(self.menu)
        else:
            self.menu_added_console = False

    def cmd_help(self, args):
        '''help commands'''
        if len(args) < 1:
            self.print_usage()
            return

        if args[0] == "about":
            print(self.about_string())
        elif args[0] == "site":
            print("See https://ardupilot.org/mavproxy/index.html for documentation")
        else:
            self.print_usage()

    def mavlink_packet(self, m):
        '''handle and incoming mavlink packets'''

    def print_usage(self):
        print("usage: mavhelp <about|site>")

def init(mpstate):
    '''initialise module'''
    return HelpModule(mpstate)



#!/usr/bin/env python
'''
access satellite map tile database

some functions are based on code from mapUtils.py in gmapcatcher

Andrew Tridgell
May 2012
released under GNU GPL v3 or later

conventions:
 - lat,lon is top left corner of area
 - ground_width is width at top of area

see also:
  https://wiki.openstreetmap.org/wiki/Zoom_levels
'''


import collections
import errno
import hashlib
import sys
import math
import threading
import os
import string
import time
import cv2
import numpy as np

from math import log, tan, radians, degrees, sin, cos, exp, pi, asin, atan

if sys.version_info.major < 3:
    from urllib2 import Request as url_request
    from urllib2 import urlopen as url_open
    from urllib2 import URLError as url_error
else:
    from urllib.request import Request as url_request
    from urllib.request import urlopen as url_open
    from urllib.error import URLError as actual_url_error
    from http.client import RemoteDisconnected
    url_error = (RemoteDisconnected, actual_url_error)

from MAVProxy.modules.lib import mp_util

class TileException(Exception):
    '''tile error class'''
    def __init__(self, msg):
        Exception.__init__(self, msg)

TILE_SERVICES = {
    # thanks to http://go2log.com/2011/09/26/fetching-tiles-for-offline-map/
    # for the URL mapping info
    "GoogleSat"      : "https://khm${GOOG_DIGIT}.google.com/kh/v=812&hl=pt-PT&x=${X}&y=${Y}&z=${ZOOM}&s=${GALILEO}",
    "GoogleMap"      : "https://mt${GOOG_DIGIT}.google.com/vt/lyrs=m@132&hl=pt-PT&x=${X}&y=${Y}&z=${ZOOM}&s=${GALILEO}",
    "GoogleTer"      : "https://mt${GOOG_DIGIT}.google.com/vt/v=t@132,r@249&hl=pt-PT&x=${X}&y=${Y}&z=${ZOOM}&s=${GALILEO}",
    "GoogleChina"    : "http://mt${GOOG_DIGIT}.google.cn/vt/lyrs=m@121&hl=en&gl=cn&x=${X}&y=${Y}&z=${ZOOM}&s=${GALILEO}",
    "MicrosoftBrMap" : "http://imakm${MS_DIGITBR}.maplink3.com.br/maps.ashx?v=${QUAD}|t&call=2.2.4",
    "MicrosoftHyb"   : "http://ecn.t${MS_DIGIT}.tiles.virtualearth.net/tiles/h${QUAD}.png?g=441&mkt=en-us&n=z",
    "MicrosoftSat"   : "http://ecn.t${MS_DIGIT}.tiles.virtualearth.net/tiles/a${QUAD}.png?g=441&mkt=en-us&n=z",
    "MicrosoftMap"   : "http://ecn.t${MS_DIGIT}.tiles.virtualearth.net/tiles/r${QUAD}.png?g=441&mkt=en-us&n=z",
    "MicrosoftTer"   : "http://ecn.t${MS_DIGIT}.tiles.virtualearth.net/tiles/r${QUAD}.png?g=441&mkt=en-us&shading=hill&n=z",
    "OpenStreetMap"  : "http://tile.openstreetmap.org/${ZOOM}/${X}/${Y}.png",
    "OSMARender"     : "http://tah.openstreetmap.org/Tiles/tile/${ZOOM}/${X}/${Y}.png",
    "OpenAerialMap"  : "http://tile.openaerialmap.org/tiles/?v=mgm&layer=openaerialmap-900913&x=${X}&y=${Y}&zoom=${OAM_ZOOM}",
    "OpenCycleMap"   : "http://andy.sandbox.cloudmade.com/tiles/cycle/${ZOOM}/${X}/${Y}.png",
    "Eniro DK,NO,SE,FI,PL" : "http://map.eniro.com/geowebcache/service/tms1.0.0/map/${ZOOM}/${X}/${ENI_Y}.png",
    "StatKartTopo2" : "http://opencache.statkart.no/gatekeeper/gk/gk.open_gmaps?layers=topo4&zoom=${ZOOM}&x=${X}&y=${Y}"

    }

# these are the md5sums of "unavailable" tiles
BLANK_TILES = set(["d16657bbee25d7f15c583f5c5bf23f50",
                    "c0e76e6e90ff881da047c15dbea380c7",
                    "d41d8cd98f00b204e9800998ecf8427e"])

# all tiles are 256x256
TILES_WIDTH = 256
TILES_HEIGHT = 256

class TileServiceInfo:
    '''a lookup object for the URL templates'''
    def __init__(self, x, y, zoom):
        self.X = x
        self.Y = y
        self.Z = zoom
        quadcode = ''
        for i in range(zoom - 1, -1, -1):
            quadcode += str((((((y >> i) & 1) << 1) + ((x >> i) & 1))))
        self.ZOOM = zoom
        self.QUAD = quadcode
        self.OAM_ZOOM = 17 - zoom
        self.GOOG_DIGIT = (x + y) & 3
        self.MS_DIGITBR = (((y & 1) << 1) + (x & 1)) + 1
        self.MS_DIGIT = (((y & 3) << 1) + (x & 1))
        self.Y_DIGIT = (x + y + zoom) % 3 + 1
        self.GALILEO = "Galileo"[0:(3 * x + y) & 7]
        self.ENI_Y = (1<<zoom)-1-y

    def __getitem__(self, a):
        return str(getattr(self, a))


class TileInfo:
    '''description of a tile'''
    def __init__(self, tile, zoom, service, offset=(0,0)):
        self.tile = tile
        (self.x, self.y) = tile
        self.zoom = zoom
        self.service = service
        (self.offsetx, self.offsety) = offset
        self.refresh_time()

    def key(self):
        '''tile cache key'''
        return (self.tile, self.zoom, self.service)

    def refresh_time(self):
        '''reset the request time'''
        self.request_time = time.time()

    def coord(self, offset=(0,0)):
        '''return lat,lon within a tile given (offsetx,offsety)'''
        (tilex, tiley) = self.tile
        (offsetx, offsety) = offset
        world_tiles = 1<<self.zoom
        x = ( tilex + 1.0*offsetx/TILES_WIDTH ) / (world_tiles/2.) - 1
        y = ( tiley + 1.0*offsety/TILES_HEIGHT) / (world_tiles/2.) - 1
        lon = mp_util.wrap_180(x * 180.0)
        y = exp(-y*2*pi)
        e = (y-1)/(y+1)
        lat = 180.0/pi * asin(e)
        return (lat, lon)

    def size(self):
        '''return tile size as (width,height) in meters'''
        (lat1, lon1) = self.coord((0,0))
        (lat2, lon2) = self.coord((TILES_WIDTH,0))
        width = mp_util.gps_distance(lat1, lon1, lat2, lon2)
        (lat2, lon2) = self.coord((0,TILES_HEIGHT))
        height = mp_util.gps_distance(lat1, lon1, lat2, lon2)
        return (width,height)

    def distance(self, lat, lon):
        '''distance of this tile from a given lat/lon'''
        (tlat, tlon) = self.coord((TILES_WIDTH/2,TILES_HEIGHT/2))
        return mp_util.gps_distance(lat, lon, tlat, tlon)

    def path(self):
        '''return relative path of tile image'''
        (x, y) = self.tile
        return os.path.join('%u' % self.zoom,
                    '%u' % y,
                    '%u.img' % x)

    def url(self, service):
        '''return URL for a tile'''
        if service not in TILE_SERVICES:
            raise TileException('unknown tile service %s' % service)
        url = string.Template(TILE_SERVICES[service])
        (x,y) = self.tile
        tile_info = TileServiceInfo(x, y, self.zoom)
        return url.substitute(tile_info)


class TileInfoScaled(TileInfo):
    '''information on a tile with scale information and placement'''
    def __init__(self, tile, zoom, scale, src, dst, service):
        TileInfo.__init__(self, tile, zoom, service)
        self.scale = scale
        (self.srcx, self.srcy) = src
        (self.dstx, self.dsty) = dst



class MPTile:
    '''map tile object'''
    def __init__(self, cache_path=None, download=True, cache_size=500,
             service="MicrosoftSat", tile_delay=0.3, debug=False,
             max_zoom=19, refresh_age=30*24*60*60):

        if cache_path is None:
            try:
                cache_path = os.path.join(os.environ['HOME'], '.tilecache')
            except Exception:
                if 'LOCALAPPDATA' in os.environ:
                    cache_path = os.path.join(os.environ['LOCALAPPDATA'], '.tilecache')
                else:
                    import tempfile
                    cache_path = os.path.join(tempfile.gettempdir(), '.tilecache')

        if not os.path.exists(cache_path):
            mp_util.mkdir_p(cache_path)

        self.cache_path = cache_path
        self.max_zoom = max_zoom
        self.min_zoom = 4
        self.download = download
        self.cache_size = cache_size
        self.tile_delay = tile_delay
        self.service = service
        self.debug = debug
        self.refresh_age = refresh_age

        if service not in TILE_SERVICES:
            raise TileException('unknown tile service %s' % service)

        # _download_pending is a dictionary of TileInfo objects
        self._download_pending = {}
        self._download_thread = None
        self._loading = mp_icon('loading.jpg')
        self._unavailable = mp_icon('unavailable.jpg')
        try:
            self._tile_cache = collections.OrderedDict()
        except AttributeError:
            # OrderedDicts in python 2.6 come from the ordereddict module
            # which is a 3rd party package, not in python2.6 distribution
            import ordereddict
            self._tile_cache = ordereddict.OrderedDict()

    def set_service(self, service):
        '''set tile service'''
        self.service = service

    def get_service(self):
        '''get tile service'''
        return self.service

    def get_service_list(self):
        '''return list of available services'''
        service_list = TILE_SERVICES.keys()
        return sorted(service_list)

    def set_download(self, download):
        '''set download enable'''
        self.download = download

    def coord_to_tile(self, lat, lon, zoom):
        '''convert lat/lon/zoom to a TileInfo'''
        world_tiles = 1<<zoom
        lon = mp_util.wrap_180(lon)
        x = world_tiles / 360.0 * (lon + 180.0)
        tiles_pre_radian = world_tiles / (2 * pi)
        e = sin(radians(lat))
        e = mp_util.constrain(e, -1+1.0e-15, 1-1.0e-15)
        y = world_tiles/2 + 0.5*log((1+e)/(1-e)) * (-tiles_pre_radian)
        offsetx = int((x - int(x)) * TILES_WIDTH)
        offsety = int((y - int(y)) * TILES_HEIGHT)
        return TileInfo((int(x) % world_tiles, int(y) % world_tiles), zoom, self.service, offset=(offsetx, offsety))

    def tile_to_path(self, tile):
        '''return full path to a tile'''
        return os.path.join(self.cache_path, self.service, tile.path())

    def coord_to_tilepath(self, lat, lon, zoom):
        '''return the tile ID that covers a latitude/longitude at
        a specified zoom level
        '''
        tile = self.coord_to_tile(lat, lon, zoom)
        return self.tile_to_path(tile)

    def tiles_pending(self):
        '''return number of tiles pending download'''
        return len(self._download_pending)

    def downloader(self):
        '''the download thread'''
        while self.tiles_pending() > 0:
            time.sleep(self.tile_delay)

            keys = sorted(self._download_pending.keys())

            # work out which one to download next, choosing by request_time
            tile_info = self._download_pending[keys[0]]
            for key in keys:
                if self._download_pending[key].request_time > tile_info.request_time:
                    tile_info = self._download_pending[key]

            url = tile_info.url(self.service)
            path = self.tile_to_path(tile_info)
            key = tile_info.key()

            try:
                if self.debug:
                    print("Downloading %s [%u left]" % (url, len(keys)))
                req = url_request(url)
                if url.find('google') != -1:
                    req.add_header('Referer', 'https://maps.google.com/')
                resp = url_open(req)
                headers = resp.info()
            except url_error as e:
                #print('Error loading %s' % url)
                if not key in self._tile_cache:
                    self._tile_cache[key] = self._unavailable
                self._download_pending.pop(key)
                if self.debug:
                    print("Failed %s: %s" % (url, str(e)))
                continue
            if 'content-type' not in headers or headers['content-type'].find('image') == -1:
                if not key in self._tile_cache:
                    self._tile_cache[key] = self._unavailable
                self._download_pending.pop(key)
                if self.debug:
                    print("non-image response %s" % url)
                continue
            else:
                img = resp.read()

            # see if its a blank/unavailable tile
            md5 = hashlib.md5(img).hexdigest()
            if md5 in BLANK_TILES:
                if self.debug:
                    print("blank tile %s" % url)
                    if not key in self._tile_cache:
                        self._tile_cache[key] = self._unavailable
                self._download_pending.pop(key)
                continue

            mp_util.mkdir_p(os.path.dirname(path))
            h = open(path+'.tmp','wb')
            h.write(img)
            h.close()
            try:
                os.unlink(path)
            except Exception:
                pass
            os.rename(path+'.tmp', path)
            self._download_pending.pop(key)
        self._download_thread = None

    def start_download_thread(self):
        '''start the downloader'''
        if self._download_thread:
            return
        t = threading.Thread(target=self.downloader)
        t.daemon = True
        self._download_thread = t
        t.start()

    def load_tile_lowres(self, tile):
        '''load a lower resolution tile from cache to fill in a
        map while waiting for a higher resolution tile'''
        if tile.zoom == self.min_zoom:
            return None

        # find the equivalent lower res tile
        (lat,lon) = tile.coord()

        width2 = TILES_WIDTH
        height2 = TILES_HEIGHT

        for zoom2 in range(tile.zoom-1, self.min_zoom-1, -1):
            width2 //= 2
            height2 //= 2

            if width2 == 0 or height2 == 0:
                break

            tile_info = self.coord_to_tile(lat, lon, zoom2)

            # see if its in the tile cache
            key = tile_info.key()
            if key in self._tile_cache:
                img = self._tile_cache[key]
                if np.array_equal(img, np.array(self._unavailable)):
                    continue
            else:
                path = self.tile_to_path(tile_info)
                img = cv2.imread(path)
                if img is None:
                    continue
                #cv2.rectangle(img, (0,0), (TILES_WIDTH-1,TILES_WIDTH-1), (255,0,0), 1)
                # add it to the tile cache
                self._tile_cache[key] = img
                while len(self._tile_cache) > self.cache_size:
                    self._tile_cache.popitem(0)

            # copy out the quadrant we want
            availx = min(TILES_WIDTH - tile_info.offsetx, width2)
            availy = min(TILES_HEIGHT - tile_info.offsety, height2)
            if availx != width2 or availy != height2:
                continue
            roi = img[tile_info.offsety:tile_info.offsety+height2, tile_info.offsetx:tile_info.offsetx+width2]

            # and scale it
            try:
                scaled = cv2.resize(roi, (TILES_HEIGHT,TILES_WIDTH))
            except Exception as ex:
                return None
            #cv.Rectangle(scaled, (0,0), (255,255), (0,255,0), 1)
            return scaled
        return None

    def load_tile(self, tile):
        '''load a tile from cache or tile server'''

        # see if its in the tile cache
        key = tile.key()
        if key in self._tile_cache:
            img = self._tile_cache[key]
            if np.array_equal(img, self._unavailable):
                img = self.load_tile_lowres(tile)
                if img is None:
                    img = self._unavailable
            return img


        path = self.tile_to_path(tile)
        ret = cv2.imread(path)
        if ret is not None:
            #cv2.rectangle(ret, (0,0), (TILES_WIDTH-1,TILES_WIDTH-1), (255,0,0), 1)
            # if it is an old tile, then try to refresh
            if os.path.getmtime(path) + self.refresh_age < time.time():
                try:
                        self._download_pending[key].refresh_time()
                except Exception:
                        self._download_pending[key] = tile
                self.start_download_thread()
                    
            # add it to the tile cache
            self._tile_cache[key] = ret
            while len(self._tile_cache) > self.cache_size:
                self._tile_cache.popitem(0)
            return ret

        if not self.download:
            img = self.load_tile_lowres(tile)
            if img is None:
                img = self._unavailable
            return img

        try:
            self._download_pending[key].refresh_time()
        except Exception:
            self._download_pending[key] = tile
        self.start_download_thread()

        img = self.load_tile_lowres(tile)
        if img is None:
            img = self._loading
        return img


    def scaled_tile(self, tile):
        '''return a scaled tile'''
        width = int(TILES_WIDTH / tile.scale)
        height = int(TILES_HEIGHT / tile.scale)
        full_tile = self.load_tile(tile)
        scaled_tile = cv2.resize(full_tile, (height, width))
        return scaled_tile


    def coord_from_area(self, x, y, lat, lon, width, ground_width):
        '''return (lat,lon) for a pixel in an area image
        x is pixel coord to the right from top,left
        y is pixel coord down from top left
        '''

        scale1 = mp_util.constrain(cos(radians(lat)), 1.0e-15, 1)
        pixel_width = ground_width / float(width)

        pixel_width_equator = (ground_width / float(width)) / cos(radians(lat))

        latr = radians(lat)
        y0 = abs(1.0/cos(latr) + tan(latr))
        lat2 = 2 * atan(y0 * exp(-(y * pixel_width_equator) / mp_util.radius_of_earth)) - pi/2.0
        lat2 = degrees(lat2)

        dx = pixel_width_equator * cos(radians(lat2)) * x

        (lat2,lon2) = mp_util.gps_offset(lat2, lon, dx, 0)

        return (lat2,lon2)

    def coord_to_pixel(self, lat, lon, width, ground_width, lat2, lon2):
        '''return pixel coordinate (px,py) for position (lat2,lon2)
        in an area image. Note that the results are relative to top,left
        and may be outside the image
        ground_width is with at lat,lon
        px is pixel coord to the right from top,left
        py is pixel coord down from top left
        '''

        pixel_width_equator = (ground_width / float(width)) / cos(radians(lat))
        latr = radians(lat)
        lat2r = radians(lat2)

        C = mp_util.radius_of_earth / pixel_width_equator
        y = C * (log(abs(1.0/cos(latr) + tan(latr))) - log(abs(1.0/cos(lat2r) + tan(lat2r))))
        y = int(y+0.5)

        dx = mp_util.gps_distance(lat2, lon, lat2, lon2)
        if mp_util.gps_bearing(lat2, lon, lat2, lon2) > 180:
            dx = -dx
        x = int(0.5 + dx / (pixel_width_equator * cos(radians(lat2))))
        return (x,y)

    def area_to_tile_list(self, lat, lon, width, height, ground_width, zoom=None):
        '''return a list of TileInfoScaled objects needed for
        an area of land, with ground_width in meters, and
        width/height in pixels.

        lat/lon is the top left corner. If unspecified, the
        zoom is automatically chosen to avoid having to grow
        the tiles
        '''

        pixel_width = ground_width / float(width)
        ground_height = ground_width * (height/(float(width)))
        top_right = mp_util.gps_newpos(lat, lon, 90, ground_width)
        bottom_left = mp_util.gps_newpos(lat, lon, 180, ground_height)
        ground_width_bottom = ground_width * cos(radians(lat)) / max(1.0e-15,cos(radians(bottom_left[0])))
        bottom_right = mp_util.gps_newpos(bottom_left[0], bottom_left[1], 90, ground_width_bottom)

        # choose a zoom level if not provided
        if zoom is None:
            zooms = range(self.min_zoom, self.max_zoom+1)
        else:
            zooms = [zoom]
        for zoom in zooms:
            tile_min = self.coord_to_tile(lat, lon, zoom)
            (twidth,theight) = tile_min.size()
            tile_pixel_width = twidth / float(TILES_WIDTH)
            scale = pixel_width / tile_pixel_width
            if scale >= 1.0:
                break

        scaled_tile_width = int(TILES_WIDTH / scale)
        scaled_tile_height = int(TILES_HEIGHT / scale)

        # work out the bottom right tile
        tile_max = self.coord_to_tile(bottom_right[0], bottom_right[1], zoom)

        ofsx = int(tile_min.offsetx / scale)
        ofsy = int(tile_min.offsety / scale)
        srcy = ofsy
        dsty = 0

        ret = []

        # place the tiles
        for y in range(tile_min.y, tile_max.y+1):
            srcx = ofsx
            dstx = 0
            world_tiles = 1<<zoom
            lim_x = (tile_max.x+1) % world_tiles
            x = tile_min.x
            while x != lim_x:
                if dstx < width and dsty < height:
                    ret.append(TileInfoScaled((x,y), zoom, scale,
                                                                  (srcx,srcy), (dstx,dsty), self.service))
                dstx += scaled_tile_width-srcx
                srcx = 0
                x = (x+1) % world_tiles
            dsty += scaled_tile_height-srcy
            srcy = 0
        return ret

    def area_to_image(self, lat, lon, width, height, ground_width, zoom=None, ordered=True):
        '''return an RGB image for an area of land, with ground_width
        in meters, and width/height in pixels.

        lat/lon is the top left corner. The zoom is automatically
        chosen to avoid having to grow the tiles'''

        img = np.zeros((height,width,3), np.uint8)

        tlist = self.area_to_tile_list(lat, lon, width, height, ground_width, zoom)

        # order the display by distance from the middle, so the download happens
        # close to the middle of the image first
        if ordered:
            (midlat, midlon) = self.coord_from_area(width/2, height/2, lat, lon, width, ground_width)
            tlist.sort(key=lambda d: d.distance(midlat, midlon), reverse=True)

        for t in tlist:
            scaled_tile = self.scaled_tile(t)

            w = width - t.dstx
            h = height - t.dsty
            if w > 0 and h > 0:
                scaled_tile_roi = scaled_tile[t.srcy:t.srcy+h, t.srcx:t.srcx+w]
                h = scaled_tile_roi.shape[0]
                w = scaled_tile_roi.shape[1]
                img[t.dsty:t.dsty+h, t.dstx:t.dstx+w] = scaled_tile_roi.copy()

        # return as an RGB image
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img

def mp_icon(filename):
    '''load an icon from the data directory'''
    # we have to jump through a lot of hoops to get an OpenCV image
    # when we may be in a package zip file
    try:
        import pkg_resources
        name = __name__
        if name == "__main__":
            name = "MAVProxy.modules.mavproxy_map.mp_tile"
        stream = pkg_resources.resource_stream(name, "data/%s" % filename).read()
        raw = np.fromstring(stream, dtype=np.uint8)
    except Exception:
        try:
            stream = open(os.path.join(__file__, 'data', filename)).read()
            raw = np.fromstring(stream, dtype=np.uint8)
        except Exception:
            #we're in a Windows exe, where pkg_resources doesn't work
            import pkgutil
            raw = pkgutil.get_data( 'MAVProxy', 'modules//mavproxy_map//data//' + filename)
    img = cv2.imdecode(raw, cv2.IMREAD_COLOR)
    return img


if __name__ == "__main__":

    from optparse import OptionParser
    parser = OptionParser("mp_tile.py [options]")
    parser.add_option("--lat", type='float', default=-35.362938, help="start latitude")
    parser.add_option("--lon", type='float', default=149.165085, help="start longitude")
    parser.add_option("--width", type='float', default=1000.0, help="width in meters")
    parser.add_option("--service", default="OviHybrid", help="tile service")
    parser.add_option("--zoom", default=None, type='int', help="zoom level")
    parser.add_option("--max-zoom", type='int', default=19, help="maximum tile zoom")
    parser.add_option("--delay", type='float', default=1.0, help="tile download delay")
    parser.add_option("--boundary", default=None, help="region boundary")
    parser.add_option("--debug", action='store_true', default=False, help="show debug info")
    (opts, args) = parser.parse_args()

    lat = opts.lat
    lon = opts.lon
    ground_width = opts.width

    if opts.boundary:
        boundary = mp_util.polygon_load(opts.boundary)
        bounds = mp_util.polygon_bounds(boundary)
        lat = bounds[0]+bounds[2]
        lon = bounds[1]
        ground_width = max(mp_util.gps_distance(lat, lon, lat, mp_util.wrap_180(lon+bounds[3])),
                           mp_util.gps_distance(lat, lon, lat-bounds[2], lon))
        print(lat, lon, ground_width)

    mt = MPTile(debug=opts.debug, service=opts.service,
            tile_delay=opts.delay, max_zoom=opts.max_zoom)
    if opts.zoom is None:
        zooms = range(mt.min_zoom, mt.max_zoom+1)
    else:
        zooms = [opts.zoom]
    for zoom in zooms:
        tlist = mt.area_to_tile_list(lat, lon, width=1024, height=1024,
                         ground_width=ground_width, zoom=zoom)
        print("zoom %u needs %u tiles" % (zoom, len(tlist)))
        for tile in tlist:
            mt.load_tile(tile)
        while mt.tiles_pending() > 0:
            time.sleep(2)
            print("Waiting on %u tiles" % mt.tiles_pending())
    print('Done')



"""
monitoring of heliplane
"""

import os, sys, math, time

from pymavlink import mavutil
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings

class HeliPlaneModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(HeliPlaneModule, self).__init__(mpstate, "heliplane", "HeliPlane", public=False)
        self.last_chan_check = 0

        self.update_channels()

    def get_rc_input(self, msg, chan):
        '''extract RC input value'''
        if chan <= 0:
            return -1
        return getattr(msg, 'chan%u_raw' % chan, -1)

    def get_pwm_output(self, msg, chan):
        '''extract PWM output value'''
        if chan <= 0:
            return -1
        return getattr(msg, 'servo%u_raw' % chan, -1)
    
    def mavlink_packet(self, msg):
        '''handle an incoming mavlink packet'''
        type = msg.get_type()

        master = self.master

        # add some status fields
        if type in [ 'RC_CHANNELS' ]:
            ilock = self.get_rc_input(msg, self.interlock_channel)
            if ilock <= 0:
                self.console.set_status('ILOCK', 'ILOCK:--', fg='grey', row=4)
            elif ilock >= 1800:
                self.console.set_status('ILOCK', 'ILOCK:ON', fg='red', row=4)
            else:
                self.console.set_status('ILOCK', 'ILOCK:OFF', fg='green', row=4)

            override = self.get_rc_input(msg, self.override_channel)
            if override <= 0:
                self.console.set_status('OVR', 'OVR:--', fg='grey', row=4)
            elif override >= 1800:
                self.console.set_status('OVR', 'OVR:ON', fg='red', row=4)
            else:
                self.console.set_status('OVR', 'OVR:OFF', fg='green', row=4)

            zeroi = self.get_rc_input(msg, self.zero_I_channel)
            if zeroi <= 0:
                self.console.set_status('ZEROI', 'ZEROI:--', fg='grey', row=4)
            elif zeroi >= 1800:
                self.console.set_status('ZEROI', 'ZEROI:ON', fg='red', row=4)
            else:
                self.console.set_status('ZEROI', 'ZEROI:OFF', fg='green', row=4)

            novtol = self.get_rc_input(msg, self.no_vtol_channel)
            if novtol <= 0:
                self.console.set_status('NOVTOL', 'NOVTOL:--', fg='grey', row=4)
            elif novtol >= 1800:
                self.console.set_status('NOVTOL', 'NOVTOL:ON', fg='red', row=4)
            else:
                self.console.set_status('NOVTOL', 'NOVTOL:OFF', fg='green', row=4)
                
        if type in [ 'SERVO_OUTPUT_RAW' ]:
            rsc = self.get_pwm_output(msg, self.rsc_out_channel)
            if rsc <= 0:
                self.console.set_status('RSC', 'RSC:--', fg='grey', row=4)
            elif rsc <= 1200:
                self.console.set_status('RSC', 'RSC:%u' % rsc, fg='red', row=4)
            elif rsc <= 1600:
                self.console.set_status('RSC', 'RSC:%u' % rsc, fg='orange', row=4)
            else:
                self.console.set_status('RSC', 'RSC:%u' % rsc, fg='green', row=4)

            thr = self.get_pwm_output(msg, self.fwd_thr_channel)
            if thr <= 0:
                self.console.set_status('FTHR', 'FTHR:--', fg='grey', row=4)
            elif thr <= 1100:
                self.console.set_status('FTHR', 'FTHR:%u' % thr, fg='red', row=4)
            elif thr <= 1500:
                self.console.set_status('FTHR', 'FTHR:%u' % thr, fg='orange', row=4)
            else:
                self.console.set_status('FTHR', 'FTHR:%u' % thr, fg='green', row=4)
                
        if type in [ 'RPM' ]:
            rpm = msg.rpm1
            if rpm < 1000:
                rpm_colour = 'red'
            elif rpm < 2000:
                rpm_colour = 'orange'
            else:
                rpm_colour = 'green'
            self.console.set_status('RPM', 'RPM: %u' % rpm, fg=rpm_colour, row=4)

    def update_channels(self):
        '''update which channels provide input'''
        self.interlock_channel = -1
        self.override_channel = -1
        self.zero_I_channel = -1
        self.no_vtol_channel = -1

        # output channels
        self.rsc_out_channel = 9
        self.fwd_thr_channel = 10

        for ch in range(1,16):
            option = self.get_mav_param("RC%u_OPTION" % ch, 0)
            if option == 32:
                self.interlock_channel = ch;
            elif option == 63:
                self.override_channel = ch;
            elif option == 64:
                self.zero_I_channel = ch;
            elif option == 65:
                self.override_channel = ch;
            elif option == 66:
                self.no_vtol_channel = ch;

            function = self.get_mav_param("SERVO%u_FUNCTION" % ch, 0)
            if function == 32:
                self.rsc_out_channel = ch
            if function == 70:
                self.fwd_thr_channel = ch

    def idle_task(self):
        '''run periodic tasks'''
        now = time.time()
        if now - self.last_chan_check >= 1:
            self.last_chan_check = now
            self.update_channels()

def init(mpstate):
    '''initialise module'''
    return HeliPlaneModule(mpstate)



#***************************************************************************
#                      Copyright Jaime Machuca
#***************************************************************************
# Title        : sc_SonyQX1.py
#
# Description  : This file contains a class to use the Sony QX range of cams
#                It finds a camera using SSDP discovery and returns it as an
#                object. If a camera is not found it returns an error value
#                that should be catched by the application. Initially it will
#                have support for triggering the camera, and downloading the
#                latest image file. Other functions will be added gradually.
#
# Environment  : Python 2.7 Code. Intended to be included in a Mavproxy Module
#
# Responsible  : Jaime Machuca
#
# License      : GNU GPL version 3
#
# Editor Used  : Xcode 6.1.1 (6A2008a)
#
#****************************************************************************

#****************************************************************************
# HEADER-FILES (Only those that are needed in this file)
#****************************************************************************

# System Header files and Module Headers
import os, sys, time, math, cv2, struct, fcntl
from datetime import datetime

# Module Dependent Headers
import requests, json, socket, StringIO
import xml.etree.ElementTree as ET
import urllib
from sc_ExifWriter import ExifWriter

# Own Headers
import ssdp

#****************************************************************************
# Constants
#****************************************************************************

# Target Initial Camera Values
targetShutterSpeed = 1600
targetAperture = 120
targetISOValue = "AUTO"

#****************************************************************************
# Class name       : SmartCamera_SonyQX
#
# Public Methods   : boGetLatestImage
#                    u32GetImageCounter
#                    boTakePicture
#                    boSetExposureMode
#                    boSetShutterSpeed
#                    boSetAperture
#                    boSetISO
#                    boZoomIn
#                    boZoomOut
#
# Private Methods  : __sFindInterfaceIPAddress
#                    __sFindCameraURL
#                    __sMakeCall
#                    __sSimpleCall
#****************************************************************************
class SmartCamera_SonyQX():

#****************************************************************************
#   Method Name     : __init__ Class Initializer
#
#   Description     : Initializes the class
#
#   Parameters      : u8instance        Camera Instance Number
#                     snetInterface     String containing the Network Interface
#                                       Name where we should look for the cam
#
#   Return Value    : None
#
#   Author           : Jaime Machuca, Randy Mackay
#
#****************************************************************************

    def __init__(self, u8Instance, sNetInterface):

        # record instance
        self.u8Instance = u8Instance
        self.sConfigGroup = "Camera%d" % self.u8Instance

        # background image processing variables
        self.u32ImgCounter = 0              # num images requested so far

        # latest image captured
        self.sLatestImageURL = None         # String with the URL to the latest image

        # latest image downloaded
        self.sLatestImageFilename = None    #String with the Filename for the last downloaded image
        self.sLatestFileName = None         #String with the camera file name for the last image taken

        self.vehicleLat = 0.0              # Current Vehicle Latitude
        self.vehicleLon = 0.0              # Current Vehicle Longitude
        self.vehicleHdg = 0.0              # Current Vehicle Heading
        self.vehicleAMSL = 0.0             # Current Vehicle Altitude above mean sea level

        self.vehicleRoll = 0.0              # Current Vehicle Roll
        self.vehiclePitch = 0.0              # Current Vehicle Pitch


        # Look Camera and Get URL
        self.sCameraURL = self.__sFindCameraURL(sNetInterface)
        if self.sCameraURL is None:
            print("No QX camera found, failed to open QX camera %d" % self.u8Instance)
        else:
            self.__openGeoTagLogFile()      # open geoTag Log
            self.boCameraInitialSetup()     # Setup Initial camera parameters

#****************************************************************************
#   Method Name     : __str__
#
#   Description     : Returns a human readable string name for the instance
#
#   Parameters      : none
#
#   Return Value    : String with object instance name
#
#   Author           : Randy Mackay
#
#****************************************************************************

    # __str__ - print position vector as string
    def __str__(self):
        return "SmartCameraSonyQX Object for %s" % self.sConfigGroup

#****************************************************************************
#   Method Name     : boCameraInitialSetup
#
#   Description     : Sets Initial Camera Parameters
#
#   Parameters      : None
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boCameraInitialSetup(self):
        print("Setting up Camera Initial Parameters")
        # Check if we need to do 'startRecMode'
        APIList = self.__sSimpleCall("getAvailableApiList")

        # For those cameras which need it
        if 'startRecMode' in (APIList['result'])[0]:
            print("Need to send startRecMode, sending and waiting 5 sec...")
            self.__sSimpleCall("startRecMode")
            time.sleep(1)
            print("4 sec")
            time.sleep(1)
            print("3 sec")
            time.sleep(1)
            print("2 sec")
            time.sleep(1)
            print("1 sec")
            time.sleep(1)

        # Set Postview Size to Orignial size to get real image filename
        sResponse = self.__sSimpleCall("getSupportedPostviewImageSize")
        print("%s" % sResponse)

        sResponse = self.__sSimpleCall("setPostviewImageSize", adictParams=["Original"])
        print("%s" % sResponse)

        sResponse = self.__sSimpleCall("getPostviewImageSize")
        print("%s" % sResponse)

        # Set Mode to Shutter Priority if available
        SupportedModes = self.__sSimpleCall("getSupportedExposureMode")
        if 'Shutter' in (SupportedModes['result'])[0]:
            self.boSetExposureMode("Shutter")
        #elif 'Manual' in (SupportedModes['result'])[0]:
        #    self.boSetExposureMode("Manual")
        else:
            print("Error no Shutter Priority Mode")

        # Set Target Shutter Speed
        self.boSetShutterSpeed(targetShutterSpeed)

        # Set Target ISO Value
        self.boSetISO(targetISOValue)

        #Take Confirmation Picture
        self.boTakePicture()

#****************************************************************************
#   Method Name     : boSet_GPS
#
#   Description     : Gets the GPS Position from the provided message
#
#   Parameters      : mGPSMessage       GPS Mavlink Message type
#
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSet_GPS(self, mGPSMessage):
        if mGPSMessage.get_type() == 'GLOBAL_POSITION_INT':
            (self.vehicleLat, self.vehicleLon, self.vehicleHdg, self.vehicleAMSL) = (mGPSMessage.lat*1.0e-7, mGPSMessage.lon*1.0e-7, mGPSMessage.hdg*0.01, mGPSMessage.alt*0.001)

#****************************************************************************
#   Method Name     : boSet_Attitude
#
#   Description     : Gets the vehicle attitude from the provided message
#
#   Parameters      : mAttitudeMessage  MAVlink Attitude Message type
#
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSet_Attitude(self, mAttitudeMessage):
        if mAttitudeMessage.get_type() == 'ATTITUDE':
            (self.vehicleRoll, self.vehiclePitch) = (math.degrees(mAttitudeMessage.roll), math.degrees(mAttitudeMessage.pitch))

#****************************************************************************
#   Method Name     : __geoRef_write
#
#   Description     : Writes GeoReference to file
#
#   Parameters      : sImageFileName    File name of image to be entered into the log
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************


    # Geo reference log for all the GoPro pictures
    def __geoRef_write(self, sImageFileName):
        #self.geoRef_writer.write(datetime.now().strftime('%d-%m-%Y %H:%M:%S.%f')[:-3])
        self.geoRef_writer.write(sImageFileName)
        self.geoRef_writer.write(",%f,%f,%f,%f,%f,%f" % (self.vehicleLat, self.vehicleLon, self.vehicleAMSL, self.vehicleRoll, self.vehiclePitch,self.vehicleHdg))
        self.geoRef_writer.write('\n')
        self.geoRef_writer.flush()

#****************************************************************************
#   Method Name     : __boWriteURLToLog
#
#   Description     : Writes images URL to file
#
#   Parameters      : sImageFileName    File name of image to be entered into the log
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    # Geo reference log for all the GoPro pictures
    def __boWriteURLToLog(self, sURLName):
        self.urlLog_writer.write(sURLName)
        self.urlLog_writer.write('\n')
        self.urlLog_writer.flush()
    
#****************************************************************************
#   Method Name     : get_real_Yaw
#
#   Description     : Helper method to get the real Yaw
#
#   Parameters      : yaw        Vehicle Yaw
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def get_real_Yaw(self, yaw):
        if (yaw < 0):
            return yaw+360
        return yaw

#****************************************************************************
#   Method Name     : __writeGeoRefToFile
#
#   Description     : Writes the Georeference of the image to the log. NOT SURE
#                     IF IT IS DUPLICATED FOR A GOOD REASON.
#
#   Parameters      : sImageFileName1   Image file name
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def __writeGeoRefToFile(self, sImageFileName1):
        self.__geoRef_write(sImageFileName1)

#****************************************************************************
#   Method Name     : __openGeoTagLogFile
#
#   Description     : Checks for existing log files and creates a new Log file
#                     with an incremented index number
#
#   Parameters      : None
#
#   Return Value    : None
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def __openGeoTagLogFile(self):
        
        # Verify folder exists
        if not os.path.exists('/sdcard/log'):
            os.makedirs('/sdcard/log')
 
        #Open GeoTag Log File
        i = 0
        while os.path.exists('/sdcard/log/geoRef%s.log' % i):
            #print('checking /sdcard/log/geoRef%s.log' % i)
            i += 1
        
        self.sCurrentGeoRefFilename = '/sdcard/log/geoRef%s.log' % i
        self.geoRef_writer = open('/sdcard/log/geoRef%s.log' % i, 'w', 0)
        self.geoRef_writer.write('Filename, Latitude, Longitude, Alt (AMSL), Roll, Pitch, Yaw\n')

        print('Opened GeoTag Log File with Filename: geoRef%s.log' % i)
        
        #Open URL Log File
        i = 0
        while os.path.exists('/sdcard/log/urlLog%s.log' % i):
            #print('checking /sdcard/log/urlLog%s.log' % i)
            i += 1
        
        self.sCurrentURLLogFilename = '/sdcard/log/urlLog%s.log' % i
        self.urlLog_writer = open('/sdcard/log/urlLog%s.log' % i, 'w', 0)

        print('Opened URL Log File with Filename: urlLog%s.log' % i)

#****************************************************************************
#   Method Name     : __sFindInterfaceIPAddress
#
#   Description     : Gets the IP Address of the interface name requested
#
#   Parameters      : sInterfaceName
#
#   Return Value    : String with the IP Address for the requested interface
#
#   Author           : Jaime Machuca,
#
#****************************************************************************

    def __sFindInterfaceIPAddress(self,sInterfaceName):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(
                                            s.fileno(),
                                            0x8915,  # SIOCGIFADDR
                                            struct.pack('256s', sInterfaceName[:15])
                                            )[20:24])

#****************************************************************************
#   Method Name     : __sMakeCall
#
#   Description     : Sends a json encoded command to the QX camera URL
#
#   Parameters      : sService
#                     adictPayload
#
#   Return Value    : JSON encoded string with camera response
#
#   Author           : Andrew Tridgell, Jaime Machuca
#
#****************************************************************************

    def __sMakeCall(self, sService, adictPayload):
        sURL = "%s/%s" % (self.sCameraURL, sService)
        adictHeaders = {"content-type": "application/json"}
        sData = json.dumps(adictPayload)
        sResponse = requests.post(sURL,
                                  data=sData,
                                  headers=adictHeaders).json()
        return sResponse

#****************************************************************************
#   Method Name     : __sSimpleCall
#
#   Description     : Articulates a camera service command to send to the QX
#                     camera
#
#   Parameters      : sMethod, command name as stated in Sony's API documentation
#                     sTarget, API Service type
#                     adictParams, command specific parameters (see Sony's API Documentation)
#                     u8Id, ??
#                     sVersion, API version for the command (see Sony's API Documentation)
#
#   Return Value    : JSON encoded string with camera response
#
#   Author           : Andrew Tridgell, Jaime Machuca
#
#****************************************************************************

    def __sSimpleCall(self, sMethod, sTarget="camera", adictParams=[], u8Id=1, sVersion="1.0"):
        print("Calling %s" % sMethod)
        return self.__sMakeCall(sTarget,
                              { "method" : sMethod,
                              "params" : adictParams,
                              "id"     : u8Id,
                              "version" : sVersion })

#****************************************************************************
#   Method Name     : __sFindCameraURL
#
#   Description     : Sends an SSDP request to look for a QX camera on the
#                     specified network interface
#
#   Parameters      : sInterface, String with the network interface name
#
#   Return Value    : String containing the URL for sending commands to the
#                     Camera
#
#   Author           : Andrew Tridgell, Jaime Machuca
#
#****************************************************************************

    def __sFindCameraURL(self, sInterface):
        sSSDPString = "urn:schemas-sony-com:service:ScalarWebAPI:1";
        sInterfaceIP = self.__sFindInterfaceIPAddress(sInterface)
        print ("Interface IP Address: %s" % sInterfaceIP)
        sRet = ssdp.discover(sSSDPString, if_ip=sInterfaceIP)
        if len(sRet) == 0:
            return None
        sDMS_URL = sRet[0].location

        print("Fetching DMS from %s" % sDMS_URL)
        xmlReq = requests.request('GET', sDMS_URL)

        xmlTree = ET.ElementTree(file=StringIO.StringIO(xmlReq.content))
        for xmlElem in xmlTree.iter():
            if xmlElem.tag == '{urn:schemas-sony-com:av}X_ScalarWebAPI_ActionList_URL':
                print("Found camera at %s" % xmlElem.text)
                return xmlElem.text
        return None

#****************************************************************************
#   Method Name     : boValidCameraFound
#
#   Description     : Returns weather or not a camera has been found. This
#                     should be used to try to find the camera again, or
#                     destroy the object.
#
#   Parameters      : none
#
#   Return Value    : True if camera has been found
#                     False if no camera has been found
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boValidCameraFound(self):
        print ("Checking URL at %s" % self.sCameraURL)
        if self.sCameraURL is None:
            return False

        return True

#****************************************************************************
#   Method Name     : boGetLatestImage
#
#   Description     : Downloads the latest image taken by the camera and then
#                     saves it to a file name composed by the camera instance
#                     and image number.
#
#   Parameters      : none
#
#   Return Value    : True if it was successful
#                     False if no image was downloaded
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boGetLatestImage(self):
        self.sLatestImageFilename = '%s_image_%s.jpg' % (self.sConfigGroup,self.u32ImgCounter)
        print ("Downloading, ",self.sLatestImageFilename)
        imgReq = requests.request('GET', self.sLatestImageURL)
        if imgReq is not None:
            open(self.sLatestImageFilename, 'w').write(imgReq.content)
            return True
        return False

#****************************************************************************
#   Method Name     : boGetAllSessionPictures
#
#   Description     : Downloads all the images stored in the URL log file to
#                     the companion computer.
#
#   Parameters      : none
#
#   Return Value    : True if it was successful
#                     False if no image was downloaded
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boGetAllSessionPictures(self, sLogFile):
        
        print("Picture Download started")
        file = open(self.sCurrentURLLogFilename, "r")
        

        for url in file:
            url = url.rstrip('\n')
            #now download link
            start = self.sLatestImageURL.find('DSC')
            end = self.sLatestImageURL.find('JPG', start) + 3
            filename = url[start:end]
            print("Downloading %s" % filename)

            geotagFile = open(self.sCurrentGeoRefFilename, "r")
            for line in geotagFile:
                if filename in line:
                    currFileLatitude = float(line.split(',')[1])
                    currFileLongitude = float(line.split(',')[2])
                    currFileAltitude = float(line.split(',')[3])
                    print ("%s,%f,%f,%f" % (filename, currFileLatitude, currFileLongitude, currFileAltitude))
            geotagFile.close()
            
            try:
                req = requests.request('GET', url)
                open('/sdcard/log/%s' % filename, 'w').write(req.content)
                ExifWriter.write_gps('/sdcard/log/%s' % filename, currFileLatitude, currFileLongitude, currFileAltitude)
            except Exception as ex:
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(ex).__name__, ex.args)
                print(message)
            
        file.close()

        return False
    
#****************************************************************************
#   Method Name     : sGetLatestImageFilename
#
#   Description     : Returns the filename of the last image downloaded from
#                     the camera
#
#   Parameters      : none
#
#   Return Value    : String containing the image file name
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def sGetLatestImageFilename(self):
        return self.sLatestImageFilename

#****************************************************************************
#   Method Name     : u32GetImageCounter
#
#   Description     : Returns the number of images taken
#
#   Parameters      : none
#
#   Return Value    : Integer with the number of images
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def u32GetImageCounter(self):
        return self.u32ImgCounter

#****************************************************************************
#   Method Name     : boZoomIn
#
#   Description     : Commands the camera to do a Zoom In step
#
#   Parameters      : None
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boZoomIn(self):

        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("actZoom", adictParams=["in","1shot"])

        # Check response for a successful result
        if 'result' in sResponse:
            print ("Zoomed in")
            return True

        # In case of an error, return false
        print ("Failed to Zoom")
        return False

#****************************************************************************
#   Method Name     : boZoomOut
#
#   Description     : Commands the camera to do a Zoom In step
#
#   Parameters      : None
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boZoomOut(self):

        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("actZoom", adictParams=["out","1shot"])

        # Check response for a successful result
        if 'result' in sResponse:
            print ("Zoomed out")
            return True

        # In case of an error, return false
        print ("Failed to Zoom")
        return False

#****************************************************************************
#   Method Name     : boSetExposureMode
#
#   Description     : Commands the camera to set a specific ShootingMode
#
#   Parameters      : Exposure Mode String
#                     Program Auto, Aperture, Shutter, Manual, Intelligent Auto, Superior Auto
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSetExposureMode(self,sExposureMode):
        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("setExposureMode", adictParams=[sExposureMode])

        # Check response for a successful result
        if 'result' in sResponse:
            time.sleep(0.25)
            sResponse = self.__sSimpleCall("getExposureMode")

            if sExposureMode not in sResponse["result"]:
                print ("Failed to set Exposure Mode, current value: %s" %sResponse["result"])
                return False

            print ("Exposure Mode set to %s" % sExposureMode)
            return True

        # In case of an error, return false
        print ("Failed to set Exposure Mode")
        return False

#****************************************************************************
#   Method Name     : boSetShutterSpeed
#
#   Description     : Commands the camera to set the Shutter Speed
#
#   Parameters      : Integer with the shutter speed divisor
#                     i.e. 1/1000 = 1000
#                     NOTE: This will only work for shutter speeds smaller than 1 sec
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSetShutterSpeed(self,u16ShutterSpeed):
        # Create Shutter Speed String
        sShutterSpeed = "1/%s" % str(u16ShutterSpeed)

        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("setShutterSpeed", adictParams=[sShutterSpeed])

        # Check response for a successful result
        if 'result' in sResponse:
            time.sleep(0.25)
            sResponse = self.__sSimpleCall("getShutterSpeed")

            if sShutterSpeed not in sResponse["result"]:
                print ("Failed to set Shutter Speed, current value: %s" %sResponse["result"])
                return False

            print ("Shutter Speed set to %s" % sShutterSpeed)
            return True

        # In case of an error, return false
        print ("Failed to set Shutter Speed")
        return False

#****************************************************************************
#   Method Name     : boSetAperture
#
#   Description     : Commands the camera to set a lens Apperture
#
#   Parameters      : F number * 10
#                     i.e. F 2.8 = 28
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSetAperture(self,u8Aperture):
        # Create Aperture String (cast one of the numbers to float to get a float)
        fFvalue = u8Aperture / float(10)
        sFValue = str(fFvalue)

        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("setFNumber", adictParams=[sFValue])

        # Check response for a successful result
        if 'result' in sResponse:
            time.sleep(0.25)
            sResponse = self.__sSimpleCall("getFNumber")

            if sFValue not in sResponse["result"]:
                print ("Failed to set aperture, current value: %s" %sResponse["result"])
                return False

            print ("Aperture set to %s" % sFValue)
            return True

        # In case of an error, return false
        print ("Failed to set aperture")
        return False

#****************************************************************************
#   Method Name     : boSetISO
#
#   Description     : Commands the camera to set an ISO number
#
#   Parameters      : ISO Value
#                     80, 100, 1000, 3200, etc...
#
#   Return Value    : True if successful
#                     False if Error Received
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boSetISO(self,u16ISO):
        # Create ISO String
        sISO = str(u16ISO)

        # Send command to set Exposure Mode
        sResponse = self.__sSimpleCall("setIsoSpeedRate", adictParams=[sISO])

        # Check response for a successful result
        if 'result' in sResponse:
            sResponse = self.__sSimpleCall("getIsoSpeedRate")

            if sISO not in sResponse["result"]:
                print ("Failed to Set ISO, current value: %s" %sResponse["result"])
                return False

            print ("ISO set to %s" % sISO)
            return True

        # In case of an error, return false
        print ("Failed to Set ISO")
        return False

#****************************************************************************
#   Method Name     : __vAddGeotagToLog
#
#   Description     : Adds an entry to the log file with the name of the image
#                     and geoposition and orientation of the shot.
#
#   Parameters      : Image file name, position, orientation
#
#   Return Value    : True if successful
#                     False if no URL was received for the image
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def __boAddGeotagToLog(self, sImageFileName):
        self.__writeGeoRefToFile(sImageFileName)
        return True

#****************************************************************************
#   Method Name     : boTakePicture
#
#   Description     : Commands the camera to take a picture
#
#   Parameters      : none
#
#   Return Value    : True if successful
#                     False if no URL was received for the image
#
#   Author           : Jaime Machuca
#
#****************************************************************************

    def boTakePicture(self):
        # Send command to take picture to camera
        sResponse = self.__sSimpleCall("actTakePicture")

        # Check response for a successful result and save latest image URL
        if 'result' in sResponse:
            self.sLatestImageURL = sResponse['result'][0][0]

            start = self.sLatestImageURL.find('DSC')
            end = self.sLatestImageURL.find('JPG', start) + 3
            self.sLatestFileName = self.sLatestImageURL[start:end]
            print("image URL: %s" % self.sLatestImageURL)
            print("image Name: %s" % self.sLatestFileName)
            self.__boAddGeotagToLog(self.sLatestFileName)
            self.__boWriteURLToLog(self.sLatestImageURL)

            self.u32ImgCounter = self.u32ImgCounter+1
            return True

        # In case of an error, return false
        return False

#****************************************************************************
#
#   Stuff Needed for testing and compatibility with current code.
#
#****************************************************************************
    def take_picture(self):
        return self.boTakePicture()

    def get_latest_image(self):
        self.boGetLatestImage()

        # this reads the image from the filename, parameter is 1 color, 0 BW, -1 unchanged
        return cv2.imread(self.sLatestImageFilename,1)


    # main - tests SmartCameraWebCam class
    def main(self):

        while True:
            # send request to image capture for image
            if self.take_picture():
                # display image
                cv2.imshow ('image_display', self.get_latest_image())
            else:
                print("no image")

            # check for ESC key being pressed
            k = cv2.waitKey(5) & 0xFF
            if k == 27:
                break

            # take a rest for a bit
            time.sleep(0.01)

# run test run from the command line
if __name__ == "__main__":
    sc_SonyQX1_0 = SmartCameraSonyQX1(0)
    sc_SonyQX1_0.main()



"""
This module adapted ANUGA
https://anuga.anu.edu.au/

------------

Implementation of Redfearn's formula to compute UTM projections from latitude and longitude

Based in part on spreadsheet
www.icsm.gov.au/gda/gdatm/redfearn.xls
downloaded from INTERGOVERNMENTAL COMMITTEE ON SURVEYING & MAPPING (ICSM)
http://www.icsm.gov.au/icsm/

"""
from MAVProxy.modules.lib.ANUGA.geo_reference import Geo_reference, DEFAULT_ZONE


def degminsec2decimal_degrees(dd,mm,ss):
    assert abs(mm) == mm
    assert abs(ss) == ss

    if dd < 0:
        sign = -1
    else:
        sign = 1

    return sign * (abs(dd) + mm/60. + ss/3600.)

def decimal_degrees2degminsec(dec):

    if dec < 0:
        sign = -1
    else:
        sign = 1

    dec = abs(dec)
    dd = int(dec)
    f = dec-dd

    mm = int(f*60)
    ss = (f*60-mm)*60

    return sign*dd, mm, ss

def redfearn(lat, lon, false_easting=None, false_northing=None,
             zone=None, central_meridian=None, scale_factor=None):
    """Compute UTM projection using Redfearn's formula

    lat, lon is latitude and longitude in decimal degrees

    If false easting and northing are specified they will override
    the standard

    If zone is specified reproject lat and long to specified zone instead of
    standard zone

    If meridian is specified, reproject lat and lon to that instead of zone. In this case
    zone will be set to -1 to indicate non-UTM projection

    Note that zone and meridian cannot both be specifed
    """


    from math import pi, sqrt, sin, cos, tan



    #GDA Specifications
    a = 6378137.0                       #Semi major axis
    inverse_flattening = 298.257222101  #1/f
    if scale_factor is None:
        K0 = 0.9996                         #Central scale factor
    else:
        K0 = scale_factor
    #print('scale', K0)
    zone_width = 6                      #Degrees

    longitude_of_central_meridian_zone0 = -183
    longitude_of_western_edge_zone0 = -186

    if false_easting is None:
        false_easting = 500000

    if false_northing is None:
        if lat < 0:
            false_northing = 10000000  #Southern hemisphere
        else:
            false_northing = 0         #Northern hemisphere)


    #Derived constants
    f = 1.0/inverse_flattening
    b = a*(1-f)       #Semi minor axis

    e2 = 2*f - f*f#    = f*(2-f) = (a^2-b^2/a^2   #Eccentricity
    e = sqrt(e2)
    e2_ = e2/(1-e2)   # = (a^2-b^2)/b^2 #Second eccentricity
    e_ = sqrt(e2_)
    e4 = e2*e2
    e6 = e2*e4

    #Foot point latitude
    n = (a-b)/(a+b) #Same as e2 - why ?
    n2 = n*n
    n3 = n*n2
    n4 = n2*n2

    G = a*(1-n)*(1-n2)*(1+9*n2/4+225*n4/64)*pi/180


    phi = lat*pi/180     #Convert latitude to radians

    sinphi = sin(phi)
    sin2phi = sin(2*phi)
    sin4phi = sin(4*phi)
    sin6phi = sin(6*phi)

    cosphi = cos(phi)
    cosphi2 = cosphi*cosphi
    cosphi3 = cosphi*cosphi2
    cosphi4 = cosphi2*cosphi2
    cosphi5 = cosphi*cosphi4
    cosphi6 = cosphi2*cosphi4
    cosphi7 = cosphi*cosphi6
    cosphi8 = cosphi4*cosphi4

    t = tan(phi)
    t2 = t*t
    t4 = t2*t2
    t6 = t2*t4

    #Radius of Curvature
    rho = a*(1-e2)/(1-e2*sinphi*sinphi)**1.5
    nu = a/(1-e2*sinphi*sinphi)**0.5
    psi = nu/rho
    psi2 = psi*psi
    psi3 = psi*psi2
    psi4 = psi2*psi2



    #Meridian distance

    A0 = 1 - e2/4 - 3*e4/64 - 5*e6/256
    A2 = 3.0/8*(e2+e4/4+15*e6/128)
    A4 = 15.0/256*(e4+3*e6/4)
    A6 = 35*e6/3072

    term1 = a*A0*phi
    term2 = -a*A2*sin2phi
    term3 = a*A4*sin4phi
    term4 = -a*A6*sin6phi

    m = term1 + term2 + term3 + term4 #OK

    if zone is not None and central_meridian is not None:
        msg = 'You specified both zone and central_meridian. Provide only one of them'
        raise ValueError(msg)

    # Zone
    if zone is None:
        zone = int((lon - longitude_of_western_edge_zone0)/zone_width)

    # Central meridian
    if central_meridian is None:
        central_meridian = zone*zone_width+longitude_of_central_meridian_zone0
    else:
        zone = -1

    omega = (lon-central_meridian)*pi/180 #Relative longitude (radians)
    omega2 = omega*omega
    omega3 = omega*omega2
    omega4 = omega2*omega2
    omega5 = omega*omega4
    omega6 = omega3*omega3
    omega7 = omega*omega6
    omega8 = omega4*omega4

    #Northing
    term1 = nu*sinphi*cosphi*omega2/2
    term2 = nu*sinphi*cosphi3*(4*psi2+psi-t2)*omega4/24
    term3 = nu*sinphi*cosphi5*\
            (8*psi4*(11-24*t2)-28*psi3*(1-6*t2)+\
             psi2*(1-32*t2)-psi*2*t2+t4-t2)*omega6/720
    term4 = nu*sinphi*cosphi7*(1385-3111*t2+543*t4-t6)*omega8/40320
    northing = false_northing + K0*(m + term1 + term2 + term3 + term4)

    #Easting
    term1 = nu*omega*cosphi
    term2 = nu*cosphi3*(psi-t2)*omega3/6
    term3 = nu*cosphi5*(4*psi3*(1-6*t2)+psi2*(1+8*t2)-2*psi*t2+t4)*omega5/120
    term4 = nu*cosphi7*(61-479*t2+179*t4-t6)*omega7/5040
    easting = false_easting + K0*(term1 + term2 + term3 + term4)

    return zone, easting, northing



def convert_from_latlon_to_utm(points=None,
                               latitudes=None,
                               longitudes=None,
                               false_easting=None,
                               false_northing=None):
    """Convert latitude and longitude data to UTM as a list of coordinates.


    Input

    points: list of points given in decimal degrees (latitude, longitude) or
    latitudes: list of latitudes   and
    longitudes: list of longitudes
    false_easting (optional)
    false_northing (optional)

    Output

    points: List of converted points
    zone:   Common UTM zone for converted points


    Notes

    Assume the false_easting and false_northing are the same for each list.
    If points end up in different UTM zones, an ANUGAerror is thrown.
    """

    old_geo = Geo_reference()
    utm_points = []
    if points is None:
        assert len(latitudes) == len(longitudes)
        points =  map(None, latitudes, longitudes)

    for point in points:

        zone, easting, northing = redfearn(float(point[0]),
                                           float(point[1]),
                                           false_easting=false_easting,
                                           false_northing=false_northing)
        new_geo = Geo_reference(zone)
        old_geo.reconcile_zones(new_geo)
        utm_points.append([easting, northing])

    return utm_points, old_geo.get_zone()



#!/usr/bin/env python
'''Module to add or clear kml layers on the map

Copyright Stephen Dade 2016
Released under the GNU GPL version 3 or later

'''

import time, math, random
from pymavlink import mavutil, mavwp

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib.mp_settings import MPSetting
from MAVProxy.modules.mavproxy_map import mp_slipmap
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import kmlread

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

class KmlReadModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(KmlReadModule, self).__init__(mpstate, "kmlread", "Add kml or kmz layers to map")
        self.add_command('kml', self.cmd_param, "kml map handling",
                         ["<clear|snapwp|snapfence>",
                          "<load> (FILENAME)", '<layers>'])
                          
        #allayers is all the loaded layers (slimap objects)
        #curlayers is only the layers displayed (checked) on the map (text list of layer names)
        self.allayers = []
        self.curlayers = []
        self.alltextlayers = []
        self.curtextlayers = []
        self.menu_added_map = False
        self.menu_needs_refreshing = True
        
        #the fence manager
        self.fenceloader = mavwp.MAVFenceLoader()
        self.snap_points = []
        
        #make the initial map menu
        if mp_util.has_wxpython:
            self.menu_fence = MPMenuSubMenu('Set Geofence', items=[])
            self.menu = MPMenuSubMenu('KML Layers',
                                  items=[MPMenuItem('Clear', 'Clear', '# kml clear')])
                          
    def cmd_param(self, args):
        '''control kml reading'''
        usage = "Usage: kml <clear | load (filename) | layers | toggle (layername) | fence (layername)> | snapfence | snapwp"
        if len(args) < 1:
            print(usage)
            return
        elif args[0] == "clear":
            self.clearkml()
        elif args[0] == "snapwp":
            self.cmd_snap_wp(args[1:])
        elif args[0] == "snapfence":
            self.cmd_snap_fence(args[1:])
        elif args[0] == "load":
            if len(args) != 2:
                print("usage: kml load <filename>")
                return
            self.loadkml(args[1])
        elif args[0] == "layers":
            for layer in self.curlayers:
                print("Found layer: " + layer)
        elif args[0] == "toggle":
            self.togglekml(args[1])
        elif args[0] == "fence":
            self.fencekml(args[1])
        else:
            print(usage)
            return

    def cmd_snap_wp(self, args):
        '''snap waypoints to KML'''
        threshold = 10.0
        if len(args) > 0:
            threshold = float(args[0])
        wpmod = self.module('wp')
        wploader = wpmod.wploader
        changed = False
        for i in range(1,wploader.count()):
            w = wploader.wp(i)
            if not wploader.is_location_command(w.command):
                continue
            lat = w.x
            lon = w.y
            best = None
            best_dist = (threshold+1)*3
            for (snap_lat,snap_lon) in self.snap_points:
                dist = mp_util.gps_distance(lat, lon, snap_lat, snap_lon)
                if dist < best_dist:
                    best_dist = dist
                    best = (snap_lat, snap_lon)
            if best is not None and best_dist <= threshold:
                if w.x != best[0] or w.y != best[1]:
                    w.x = best[0]
                    w.y = best[1]
                    print("Snapping WP %u to %f %f" % (i, w.x, w.y))
                    wploader.set(w, i)
                    changed = True
            elif best is not None:
                if best_dist <= (threshold+1)*3:
                    print("Not snapping wp %u dist %.1f" % (i, best_dist))
        if changed:
            wpmod.send_all_waypoints()

    def cmd_snap_fence(self, args):
        '''snap fence to KML'''
        threshold = 10.0
        if len(args) > 0:
            threshold = float(args[0])
        fencemod = self.module('fence')
        loader = fencemod.fenceloader
        changed = False
        for i in range(0,loader.count()):
            fp = loader.point(i)
            lat = fp.lat
            lon = fp.lng
            best = None
            best_dist = (threshold+1)*3
            for (snap_lat,snap_lon) in self.snap_points:
                dist = mp_util.gps_distance(lat, lon, snap_lat, snap_lon)
                if dist < best_dist:
                    best_dist = dist
                    best = (snap_lat, snap_lon)
            if best is not None and best_dist <= threshold:
                if best[0] != lat or best[1] != lon:
                    loader.move(i, best[0], best[1])
                    print("Snapping fence point %u to %f %f" % (i, best[0], best[1]))
                    changed = True
            elif best is not None:
                if best_dist <= (threshold+1)*3:
                    print("Not snapping fence point %u dist %.1f" % (i, best_dist))

        if changed:
            fencemod.send_fence()

    def fencekml(self, layername):
        '''set a layer as the geofence'''
        #Strip quotation marks if neccessary
        if layername.startswith('"') and layername.endswith('"'):
            layername = layername[1:-1]
        
        #for each point in the layer, add it in
        for layer in self.allayers:
            if layer.key == layername:
                #clear the current fence
                self.fenceloader.clear()
                if len(layer.points) < 3:
                    return
                self.fenceloader.target_system = self.target_system
                self.fenceloader.target_component = self.target_component
                #send centrepoint  to fence[0] as the return point
                bounds = mp_util.polygon_bounds(layer.points)
                (lat, lon, width, height) = bounds
                center = (lat+width/2, lon+height/2)
                self.fenceloader.add_latlon(center[0], center[1])
                for lat, lon in layer.points:
                    #add point
                    self.fenceloader.add_latlon(lat, lon)
                #and send
                self.send_fence()

    def send_fence(self):
        '''send fence points from fenceloader. Taken from fence module'''
        # must disable geo-fencing when loading
        self.fenceloader.target_system = self.target_system
        self.fenceloader.target_component = self.target_component
        self.fenceloader.reindex()
        action = self.get_mav_param('FENCE_ACTION', mavutil.mavlink.FENCE_ACTION_NONE)
        self.param_set('FENCE_ACTION', mavutil.mavlink.FENCE_ACTION_NONE, 3)
        self.param_set('FENCE_TOTAL', self.fenceloader.count(), 3)
        for i in range(self.fenceloader.count()):
            p = self.fenceloader.point(i)
            self.master.mav.send(p)
            p2 = self.fetch_fence_point(i)
            if p2 is None:
                self.param_set('FENCE_ACTION', action, 3)
                return False
            if (p.idx != p2.idx or
                abs(p.lat - p2.lat) >= 0.00003 or
                abs(p.lng - p2.lng) >= 0.00003):
                print("Failed to send fence point %u" % i)
                self.param_set('FENCE_ACTION', action, 3)
                return False
        self.param_set('FENCE_ACTION', action, 3)
        return True

    def fetch_fence_point(self ,i):
        '''fetch one fence point. Taken from fence module'''
        self.master.mav.fence_fetch_point_send(self.target_system,
                                                    self.target_component, i)
        tstart = time.time()
        p = None
        while time.time() - tstart < 3:
            p = self.master.recv_match(type='FENCE_POINT', blocking=False)
            if p is not None:
                break
            time.sleep(0.1)
            continue
        if p is None:
            self.console.error("Failed to fetch point %u" % i)
            return None
        return p
                                    
    def togglekml(self, layername):
        '''toggle the display of a kml'''
        #Strip quotation marks if neccessary
        if layername.startswith('"') and layername.endswith('"'):
            layername = layername[1:-1]
        #toggle layer off (plus associated text element)
        if layername in self.curlayers:
            for layer in self.curlayers:
                if layer == layername:
                    self.mpstate.map.remove_object(layer)
                    self.curlayers.remove(layername)
                    if layername in self.curtextlayers:
                        for clayer in self.curtextlayers:
                            if clayer == layername:
                                self.mpstate.map.remove_object(clayer)
                                self.curtextlayers.remove(clayer)
        #toggle layer on (plus associated text element)
        else:
            for layer in self.allayers:
                if layer.key == layername:
                    self.mpstate.map.add_object(layer)
                    self.curlayers.append(layername)
                    for alayer in self.alltextlayers:
                        if alayer.key == layername:
                            self.mpstate.map.add_object(alayer)
                            self.curtextlayers.append(layername)
        self.menu_needs_refreshing = True
        
    def clearkml(self):
        '''Clear the kmls from the map'''
        #go through all the current layers and remove them
        for layer in self.curlayers:
            self.mpstate.map.remove_object(layer)
        for layer in self.curtextlayers:
            self.mpstate.map.remove_object(layer)
        self.allayers = []
        self.curlayers = []
        self.alltextlayers = []
        self.curtextlayers = []
        self.menu_needs_refreshing = True
                
    def loadkml(self, filename):
        '''Load a kml from file and put it on the map'''
        #Open the zip file
        nodes = kmlread.readkmz(filename)

        self.snap_points = []

        #go through each object in the kml...
        if nodes is None:
            print("No nodes found")
            return
        for n in nodes:
            try:
                point = kmlread.readObject(n)
            except Exception as ex:
                continue

            #and place any polygons on the map
            if self.mpstate.map is not None and point[0] == 'Polygon':
                self.snap_points.extend(point[2])

                #print("Adding " + point[1])
                newcolour = (random.randint(0, 255), 0, random.randint(0, 255))
                curpoly = mp_slipmap.SlipPolygon(point[1], point[2],
                                                             layer=2, linewidth=2, colour=newcolour)
                self.mpstate.map.add_object(curpoly)
                self.allayers.append(curpoly)
                self.curlayers.append(point[1])
                
                
            #and points - barrell image and text
            if self.mpstate.map is not None and point[0] == 'Point':
                #print("Adding " + point[1])
                icon = self.mpstate.map.icon('barrell.png')
                curpoint = mp_slipmap.SlipIcon(point[1], latlon = (point[2][0][0], point[2][0][1]), layer=3, img=icon, rotation=0, follow=False)
                curtext = mp_slipmap.SlipLabel(point[1], point = (point[2][0][0], point[2][0][1]), layer=4, label=point[1], colour=(0,255,255))
                self.mpstate.map.add_object(curpoint)
                self.mpstate.map.add_object(curtext)
                self.allayers.append(curpoint)
                self.alltextlayers.append(curtext)
                self.curlayers.append(point[1])
                self.curtextlayers.append(point[1])
        self.menu_needs_refreshing = True
        

    def idle_task(self):
        '''handle GUI elements'''
        if not self.menu_needs_refreshing:
            return
        if self.module('map') is not None and not self.menu_added_map:
            self.menu_added_map = True
            self.module('map').add_menu(self.menu)
        #(re)create the menu
        if mp_util.has_wxpython and self.menu_added_map:
            # we don't dynamically update these yet due to a wx bug
            self.menu.items = [ MPMenuItem('Clear', 'Clear', '# kml clear'), MPMenuItem('Load', 'Load', '# kml load ', handler=MPMenuCallFileDialog(flags=('open',), title='KML Load', wildcard='*.kml;*.kmz')), self.menu_fence, MPMenuSeparator() ]
            self.menu_fence.items = []
            for layer in self.allayers:
                #if it's a polygon, add it to the "set Geofence" list
                if isinstance(layer, mp_slipmap.SlipPolygon):
                    self.menu_fence.items.append(MPMenuItem(layer.key, layer.key, '# kml fence \"' + layer.key + '\"'))
                #then add all the layers to the menu, ensuring to check the active layers
                #text elements aren't included on the menu
                if layer.key in self.curlayers and layer.key[-5:] != "-text":
                    self.menu.items.append(MPMenuCheckbox(layer.key, layer.key, '# kml toggle \"' + layer.key + '\"', checked=True))
                elif layer.key[-5:] != "-text":
                    self.menu.items.append(MPMenuCheckbox(layer.key, layer.key, '# kml toggle \"' + layer.key + '\"', checked=False))
            #and add the menu to the map popu menu
            self.module('map').add_menu(self.menu)
        self.menu_needs_refreshing = False
                            
    def mavlink_packet(self, m):
        '''handle a mavlink packet'''


def init(mpstate):
    '''initialise module'''
    return KmlReadModule(mpstate)
    
    
    



#!/usr/bin/env python
'''calibration command handling'''

import time, os
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module

class CalibrationModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(CalibrationModule, self).__init__(mpstate, "calibration")
        self.add_command('ground', self.cmd_ground,   'do a ground start')
        self.add_command('level', self.cmd_level,    'set level on a multicopter')
        self.add_command('compassmot', self.cmd_compassmot, 'do compass/motor interference calibration')
        self.add_command('calpress', self.cmd_calpressure,'calibrate pressure sensors')
        self.add_command('accelcal', self.cmd_accelcal, 'do 3D accelerometer calibration')
        self.add_command('accelcalsimple', self.cmd_accelcal_simple, 'do simple accelerometer calibration')
        self.add_command('gyrocal', self.cmd_gyrocal, 'do gyro calibration')
        self.add_command('ahrstrim', self.cmd_ahrstrim, 'do AHRS trim')
        self.add_command('magcal', self.cmd_magcal, "magcal")
        self.add_command('forcecal', self.cmd_forcecal, "force calibration save")
        self.accelcal_count = -1
        self.accelcal_wait_enter = False
        self.compassmot_running = False
        self.empty_input_count = 0
        self.magcal_progess = []

    def cmd_ground(self, args):
        '''do a ground start mode'''
        self.master.calibrate_imu()

    def cmd_level(self, args):
        '''run a accel level'''
        print("level is no longer supported; use ahrstrim, accelcal or accelcalsimple")

    def cmd_accelcal(self, args):
        '''do a full 3D accel calibration'''
        mav = self.master
        # ack the APM to begin 3D calibration of accelerometers
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  0, 0, 0, 0, 1, 0, 0)
        self.accelcal_count = 0
        self.accelcal_wait_enter = False

    def cmd_accelcal_simple(self, args):
        '''do a simple accel calibration'''
        mav = self.master
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  0, 0, 0, 0, 4, 0, 0)

    def cmd_forcecal(self, args):
        '''force calibration save'''
        usage = "usage: forcecal accel|compass|both"
        if len(args) < 1:
            print(usage)
            return
        param2 = 0
        param5 = 0
        if args[0].lower() == "accel":
            param5 = 76
        elif args[0].lower() == "compass":
            param2 = 76
        elif args[0].lower() == "both":
            param2 = 76
            param5 = 76
        else:
            print(usage)
            return
        mav = self.master
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  0, param2, 0, 0, param5, 0, 0)
        
    def cmd_gyrocal(self, args):
        '''do a full gyro calibration'''
        mav = self.master
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  1, 0, 0, 0, 0, 0, 0)

    def cmd_ahrstrim(self, args):
        '''do a AHRS trim'''
        mav = self.master
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  0, 0, 0, 0, 2, 0, 0)

    def mavlink_packet(self, m):
        '''handle mavlink packets'''
        mtype = m.get_type()
        if self.accelcal_count != -1:
            if mtype == 'STATUSTEXT':
                # handle accelcal packet
                text = str(m.text)
                if text.startswith('Place '):
                    self.accelcal_wait_enter = True
                    self.empty_input_count = self.mpstate.empty_input_count
        if mtype == 'MAG_CAL_PROGRESS':
            while m.compass_id >= len(self.magcal_progess):
                self.magcal_progess.append("")
            self.magcal_progess[m.compass_id] = "%u%%" % m.completion_pct
            self.console.set_status('Progress', 'Calibration Progress: ' + " ".join(self.magcal_progess), row=4)
        elif mtype == 'MAG_CAL_REPORT':
            if m.cal_status == mavutil.mavlink.MAG_CAL_SUCCESS:
                result = "SUCCESS"
            else:
                result = "FAILED"
            self.magcal_progess[m.compass_id] = result
            self.console.set_status('Progress', 'Calibration Progress: ' + " ".join(self.magcal_progess), row=4)
            print("Calibration of compass %u %s: fitness %.3f" % (m.compass_id, result, m.fitness))
            mav = self.master
            mav.mav.command_long_send(mav.target_system, mav.target_component,
                                      mavutil.mavlink.MAV_CMD_DO_ACCEPT_MAG_CAL, 0,
                                      1<<m.compass_id, 0, 0, 0, 0, 0, 0)

    def idle_task(self):
        '''handle mavlink packets'''
        if self.accelcal_count != -1:
            if self.accelcal_wait_enter and self.empty_input_count != self.mpstate.empty_input_count:
                self.accelcal_wait_enter = False
                self.accelcal_count += 1
                # tell the APM that user has done as requested
                self.master.mav.command_ack_send(self.accelcal_count, 1)
                if self.accelcal_count >= 6:
                    self.accelcal_count = -1

        if self.compassmot_running:
            if self.mpstate.empty_input_count != self.empty_input_count:
                # user has hit enter, stop the process
                    self.compassmot_running = False
                    print("sending stop")
                    self.master.mav.command_ack_send(0, 1)


    def cmd_compassmot(self, args):
        '''do a compass/motor interference calibration'''
        mav = self.master
        print("compassmot starting")
        mav.mav.command_long_send(mav.target_system, mav.target_component,
                                  mavutil.mavlink.MAV_CMD_PREFLIGHT_CALIBRATION, 0,
                                  0, 0, 0, 0, 0, 1, 0)
        self.compassmot_running = True
        self.empty_input_count = self.mpstate.empty_input_count

    def cmd_calpressure(self, args):
        '''calibrate pressure sensors'''
        self.master.calibrate_pressure()

    def print_magcal_usage(self):
        print("Usage: magcal <start|accept|cancel|yaw>")

    def cmd_magcal(self, args):
        '''control magnetometer calibration'''
        if len(args) < 1:
            self.print_magcal_usage()
            return

        if args[0] == 'start':
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                0, # target_component
                mavutil.mavlink.MAV_CMD_DO_START_MAG_CAL, # command
                0, # confirmation
                0, # p1: mag_mask
                0, # p2: retry
                1, # p3: autosave
                0, # p4: delay
                0, # param5
                0, # param6
                0) # param7
        elif args[0] == 'accept':
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                0, # target_component
                mavutil.mavlink.MAV_CMD_DO_ACCEPT_MAG_CAL, # command
                0, # confirmation
                0, # p1: mag_mask
                0, # param2
                1, # param3
                0, # param4
                0, # param5
                0, # param6
                0) # param7
        elif args[0] == 'cancel':
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                0, # target_component
                mavutil.mavlink.MAV_CMD_DO_CANCEL_MAG_CAL, # command
                0, # confirmation
                0, # p1: mag_mask
                0, # param2
                1, # param3
                0, # param4
                0, # param5
                0, # param6
                0) # param7
        elif args[0] == 'yaw':
            if len(args) < 2:
                print("Usage: magcal yaw YAW_DEGREES <mask>")
                return
            yaw_deg = float(args[1])
            mask = 0
            if len(args) > 2:
                mask = int(args[2])
            print("Calibrating for yaw %.1f degrees with mask 0x%02x" % (yaw_deg, mask))
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                0, # target_component
                mavutil.mavlink.MAV_CMD_FIXED_MAG_CAL_YAW, # command
                0, # confirmation
                yaw_deg, # p1: yaw in degrees
                mask, # p2: mask
                0, # p3: lat_deg
                0, # p4: lon_deg
                0, # param5
                0, # param6
                0) # param7
        else:
            self.print_magcal_usage()
            return

def init(mpstate):
    '''initialise module'''
    return CalibrationModule(mpstate)
